#
# Copyright (C) 2017 Ericsson India Global Services Pvt Ltd.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.
#

from unittest import mock

from networking_odl.bgpvpn import odl_v2 as driverv2
from networking_odl.common import constants as odl_const
from networking_odl.common import odl_features
from networking_odl.db import db
from networking_odl.tests.unit import base_v2
from neutron_lib.api.definitions import bgpvpn_vni as bgpvpn_vni_def


class OpenDaylightBgpvpnDriverTestCase(base_v2.OpenDaylightConfigBase):

    def setUp(self):
        super(OpenDaylightBgpvpnDriverTestCase, self).setUp()
        self.driver = driverv2.OpenDaylightBgpvpnDriver(service_plugin=None)

    def _get_fake_bgpvpn(self, net=False, router=False):
        net_id = []
        router_id = []
        if router:
            router_id = ['ROUTER_ID']
        if net:
            net_id = ['NET_ID']
        fake_bgpvpn = {'route_targets': '100:1',
                       'route_distinguishers': ['100:1'],
                       'id': 'BGPVPN_ID',
                       'networks': net_id,
                       'routers': router_id}
        return fake_bgpvpn

    def _get_fake_router_assoc(self):
        fake_router_assoc = {'id': 'ROUTER_ASSOC_ID',
                             'bgpvpn_id': 'BGPVPN_ID',
                             'router_id': 'ROUTER_ID'}
        return fake_router_assoc

    def _get_fake_net_assoc(self):
        fake_net_assoc = {'id': 'NET_ASSOC_ID',
                          'bgpvpn_id': 'BGPVPN_ID',
                          'network_id': 'NET_ID'}
        return fake_net_assoc

    def _assert_op(self, operation, object_type, data, precommit=True):
        rows = sorted(db.get_all_db_rows_by_state(self.db_context,
                                                  odl_const.PENDING),
                      key=lambda x: x.seqnum)
        if precommit:
            self.db_context.session.flush()
            self.assertEqual(operation, rows[0]['operation'])
            self.assertEqual(object_type, rows[0]['object_type'])
            self.assertEqual(data['id'], rows[0]['object_uuid'])
        else:
            self.assertEqual([], rows)

    def test_create_bgpvpn(self):
        fake_data = self._get_fake_bgpvpn()
        self.driver.create_bgpvpn_precommit(self.db_context, fake_data)
        self._assert_op(odl_const.ODL_CREATE, odl_const.ODL_BGPVPN,
                        fake_data)
        self.run_journal_processing()
        self._assert_op(odl_const.ODL_CREATE, odl_const.ODL_BGPVPN,
                        fake_data, False)

    def test_update_bgpvpn(self):
        fake_data = self._get_fake_bgpvpn()
        self.driver.update_bgpvpn_precommit(self.db_context, fake_data)
        self._assert_op(odl_const.ODL_UPDATE, odl_const.ODL_BGPVPN,
                        fake_data)
        self.run_journal_processing()
        self._assert_op(odl_const.ODL_UPDATE, odl_const.ODL_BGPVPN,
                        fake_data, False)

    def test_delete_bgpvpn(self):
        fake_data = self._get_fake_bgpvpn()
        self.driver.delete_bgpvpn_precommit(self.db_context, fake_data)
        self._assert_op(odl_const.ODL_DELETE, odl_const.ODL_BGPVPN,
                        fake_data)
        self.run_journal_processing()
        self._assert_op(odl_const.ODL_DELETE, odl_const.ODL_BGPVPN,
                        fake_data, False)

    def test_create_router_assoc(self):
        fake_rtr_assoc_data = self._get_fake_router_assoc()
        fake_rtr_upd_bgpvpn_data = self._get_fake_bgpvpn(router=True)
        with mock.patch.object(self.driver, 'get_router_assocs',
                               return_value=[]), \
            mock.patch.object(self.driver, 'get_bgpvpn',
                              return_value=fake_rtr_upd_bgpvpn_data):
            self.driver.create_router_assoc_precommit(self.db_context,
                                                      fake_rtr_assoc_data)
            self._assert_op(odl_const.ODL_UPDATE,
                            odl_const.ODL_BGPVPN,
                            fake_rtr_upd_bgpvpn_data)
            self.run_journal_processing()
            self._assert_op(odl_const.ODL_UPDATE,
                            odl_const.ODL_BGPVPN,
                            fake_rtr_upd_bgpvpn_data, False)

    def test_delete_router_assoc(self):
        fake_rtr_assoc_data = self._get_fake_router_assoc()
        fake_bgpvpn_data = self._get_fake_bgpvpn(router=True)
        with mock.patch.object(self.driver, 'get_bgpvpn',
                               return_value=fake_bgpvpn_data):
            self.driver.delete_router_assoc_precommit(self.db_context,
                                                      fake_rtr_assoc_data)
            self._assert_op(odl_const.ODL_UPDATE,
                            odl_const.ODL_BGPVPN,
                            fake_bgpvpn_data)
            self.run_journal_processing()
            self._assert_op(odl_const.ODL_UPDATE,
                            odl_const.ODL_BGPVPN,
                            fake_bgpvpn_data, False)

    def test_create_net_assoc(self):
        fake_net_assoc_data = self._get_fake_net_assoc()
        fake_net_upd_bgpvpn_data = self._get_fake_bgpvpn(net=True)
        # todo(vivekanandan) add check for case when assoc already exists
        with mock.patch.object(self.driver, 'get_bgpvpns',
                               return_value=[fake_net_upd_bgpvpn_data]):
            self.driver.create_net_assoc_precommit(self.db_context,
                                                   fake_net_assoc_data)
            self._assert_op(odl_const.ODL_UPDATE,
                            odl_const.ODL_BGPVPN,
                            fake_net_upd_bgpvpn_data)
            self.run_journal_processing()
            self._assert_op(odl_const.ODL_UPDATE,
                            odl_const.ODL_BGPVPN,
                            fake_net_upd_bgpvpn_data, False)

    def test_delete_net_assoc(self):
        fake_net_assoc_data = self._get_fake_net_assoc()
        fake_bgpvpn_data = self._get_fake_bgpvpn(net=True)
        with mock.patch.object(self.driver, 'get_bgpvpn',
                               return_value=fake_bgpvpn_data):
            self.driver.delete_net_assoc_precommit(self.db_context,
                                                   fake_net_assoc_data)
            self._assert_op(odl_const.ODL_UPDATE,
                            odl_const.ODL_BGPVPN,
                            fake_bgpvpn_data)
            self.run_journal_processing()
            self._assert_op(odl_const.ODL_UPDATE,
                            odl_const.ODL_BGPVPN,
                            fake_bgpvpn_data, False)

    def _get_bgpvpn_driver_with_vni(self):
        feature_json = """{"features": {"feature":
                                        [{"service-provider-feature":
                                        "neutron-extensions:operational-port-status"},
                                        {"service-provider-feature":
                                        "neutron-extensions:bgpvpn-vni"}]}}"""

        self.cfg.config(odl_features_json=feature_json, group='ml2_odl')
        odl_features.init()
        bgpvpn_driver = driverv2.OpenDaylightBgpvpnDriver(service_plugin=None)
        return bgpvpn_driver

    def test_bgpvpn_vni_feature(self):
        bgpvpn_driver = self._get_bgpvpn_driver_with_vni()
        self.assertIn(bgpvpn_vni_def.ALIAS,
                      bgpvpn_driver.more_supported_extension_aliases)

    def test_bgpvpn_vni_create_with_vni(self):
        bgpvpn_driver = self._get_bgpvpn_driver_with_vni()
        fake_data = self._get_fake_bgpvpn()
        fake_data['vni'] = 100
        bgpvpn_driver.create_bgpvpn_precommit(self.db_context, fake_data)
        self._assert_op(odl_const.ODL_CREATE, odl_const.ODL_BGPVPN,
                        fake_data)
        self.run_journal_processing()
        self._assert_op(odl_const.ODL_CREATE, odl_const.ODL_BGPVPN,
                        fake_data, False)



# Copyright (c) 2017 NEC Corp.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import signal
from unittest import mock

import fixtures

from neutron.common import utils
from oslo_db import exception
from oslo_log import log as logging
from oslo_service.tests import test_service
from oslo_utils import uuidutils

from networking_odl.common import client
from networking_odl.common import constants as odl_const
from networking_odl.db import db
from networking_odl.db import models
from networking_odl.journal import cleanup
from networking_odl.journal import dependency_validations
from networking_odl.journal import full_sync
from networking_odl.journal import journal
from networking_odl.journal import periodic_task
from networking_odl.journal import recovery
from networking_odl.journal import worker
from networking_odl.tests import base
from networking_odl.tests.unit import base_v2
from networking_odl.tests.unit.db import test_db


PROCESS_RUNNING_STATUSES = ('S', 'R', 'D')


class JournalPeriodicProcessorTest(base_v2.OpenDaylightConfigBase,
                                   test_service.ServiceTestBase):
    def setUp(self):
        super(JournalPeriodicProcessorTest, self).setUp()
        self.periodic_task_fixture = self.useFixture(
            base.OpenDaylightPeriodicTaskFixture())
        self.cfg.config(sync_timeout=0.1, group='ml2_odl')

    def _create_periodic_processor(self):
        periodic_processor = worker.JournalPeriodicProcessor()
        self.addCleanup(periodic_processor.stop)
        return periodic_processor

    def _get_pid_status(self, pid):
        """Allows to query a system process based on the PID

        It will use `ps` to query the pid, it's state and the command.

        :param pid: An integer with the Process ID number
        :returns: A tuple of strings with the command and the running status
                  in a single char as defined in the manpage PS(1) under
                  PROCESS STATE CODES.
        """
        with os.popen('ps ax -o pid,state,cmd') as f:
            # Skip ps header
            f.readline()

            processes = (line.strip().split()[:3] for line in f)

            return next(((c, s) for p, s, c in processes if int(p) == pid),
                        (None, None))

    def _kill_process(self, pid):
        if self._get_pid_status(pid)[1] in PROCESS_RUNNING_STATUSES:
            os.kill(pid, signal.SIGKILL)

    def mock_object_with_ipc(self, target, attribute, pre_hook=None):
        patcher = mock.patch.object(target, attribute, autospec=True)
        c2p_read = self.create_ipc_for_mock(patcher, pre_hook)
        return c2p_read

    def create_ipc_for_mock(self, patcher, pre_hook=None):
        # NOTE(mpeterson): The following pipe is being used because this is
        # testing something inter processeses and we need to have a value on
        # the side of the test processes to know it succeeded with the
        # operation. A pipe provide a way for two processes to communicate.
        # The was_called method will be called by the worker process while
        # the test process will read the result on c2p_read.
        c2p_read, c2p_write = os.pipe()

        def close_pipe_end(fd):
            try:
                os.close(fd)
            except OSError:
                print('failed closing: %s' % fd)

        # First we want to close the write, to unlock any running read()
        self.addCleanup(close_pipe_end, c2p_read)
        self.addCleanup(close_pipe_end, c2p_write)

        mock_ = patcher.start()
        self.addCleanup(patcher.stop)

        def was_called(*args, **kwargs):
            # OSError is caught because start is called twice on the worker
            # and the second time the pipe is already closed.
            try:
                os.close(c2p_read)
                try:
                    if pre_hook:
                        pre_hook(*args, **kwargs)
                    os.write(c2p_write, b'1')
                except Exception:
                    # This is done so any read on the pipe is unblocked.
                    os.write(c2p_write, b'0')
                finally:
                    os.close(c2p_write)
            except OSError:
                pass

        mock_.side_effect = was_called

        return c2p_read

    def assert_ipc_mock_called(self, c2p_read):
        # If it timeouts on the read then it means the function was
        # not called.
        called = int(os.read(c2p_read, 1))

        self.assertEqual(called, 1,
                         'The IPC mock was called but during the '
                         'execution an exception was raised')

    @mock.patch.object(journal.OpenDaylightJournalThread, 'set_sync_event')
    def test_processing(self, mock_journal):
        periodic_processor = self._create_periodic_processor()
        periodic_processor.start()
        utils.wait_until_true(lambda: mock_journal.call_count > 1, 5, 0.1)

    @mock.patch.object(journal.OpenDaylightJournalThread, 'start')
    @mock.patch.object(journal.OpenDaylightJournalThread, 'stop')
    def test_stops_journal_sync_thread(self, mock_stop, mock_start):
        periodic_processor = self._create_periodic_processor()
        periodic_processor.start()
        periodic_processor.stop()
        mock_stop.assert_called_once()
        mock_start.assert_called_once()

    def test_allow_multiple_starts_gracefully(self):
        periodic_processor = self._create_periodic_processor()
        periodic_processor.start()
        periodic_processor.stop()

        try:
            periodic_processor.start()
        except RuntimeError:
            self.fail('Calling a start() after a stop() should be allowed')

    def test_multiple_starts_without_stop_throws_exception(self):
        periodic_processor = self._create_periodic_processor()
        periodic_processor.start()

        self.assertRaises(RuntimeError, periodic_processor.start)

    def test_call_stop_without_calling_start(self):
        periodic_processor = self._create_periodic_processor()

        try:
            periodic_processor.stop()
        except AttributeError:
            self.fail('start() was not called before calling stop()')

    def assert_process_running(self, pid):
        cmd, state = self._get_pid_status(pid)
        self.assertIn(state, PROCESS_RUNNING_STATUSES)
        return cmd

    def _create_periodic_processor_ipc_fork(self, target, pre_hook=None):
        self._setup_mocks_for_periodic_task()

        real_start = worker.JournalPeriodicProcessor.start
        pipe_start = self.mock_object_with_ipc(worker.JournalPeriodicProcessor,
                                               'start', real_start)

        c2p_read = self.mock_object_with_ipc(worker.JournalPeriodicProcessor,
                                             target, pre_hook)

        pid = self._spawn_service(
            service_maker=lambda: worker.JournalPeriodicProcessor())
        self.addCleanup(self._kill_process, pid)

        # Allow the process to spawn and signal handling to be registered
        self.assert_ipc_mock_called(pipe_start)

        return pid, c2p_read

    @mock.patch.object(periodic_task.PeriodicTask, 'execute_ops',
                       new=mock.Mock())
    @mock.patch.object(journal.OpenDaylightJournalThread,
                       'sync_pending_entries', new=mock.Mock())
    def test_handle_sighup_gracefully(self):
        real_reset = worker.JournalPeriodicProcessor.reset
        pid, c2p_read = self._create_periodic_processor_ipc_fork('reset',
                                                                 real_reset)

        cmd = self.assert_process_running(pid)

        os.kill(pid, signal.SIGHUP)

        self.assert_ipc_mock_called(c2p_read)

        new_cmd = self.assert_process_running(pid)
        self.assertEqual(cmd, new_cmd)

    def _setup_mocks_for_periodic_task(self, executed_recently=False):
        mock_db_module = mock.MagicMock(spec=db)
        mock_db_module.was_periodic_task_executed_recently.return_value = \
            executed_recently
        mock_db = mock.patch('networking_odl.journal.periodic_task.db',
                             mock_db_module)
        mock_db.start()
        self.addCleanup(mock_db.stop)

    @mock.patch.object(cleanup, 'delete_completed_rows')
    @mock.patch.object(cleanup, 'cleanup_processing_rows')
    @mock.patch.object(full_sync, 'full_sync')
    @mock.patch.object(recovery, 'journal_recovery')
    # ^^ The above mocks represent the required calling order starting from
    # top. Use decorators *only* to specify the stack order.
    def test_maintenance_task_correctly_registered(self, *stack_order):
        calls = []
        for item in reversed(stack_order):
            calls.append(mock.call(item))

        with mock.patch.object(
                periodic_task.PeriodicTask,
                'register_operation') as register_operation_mock:
            periodic_processor = self._create_periodic_processor()
            periodic_processor._start_maintenance_task()
            register_operation_mock.assert_has_calls(calls)

    def test_maintenance_task_started(self):
        self.periodic_task_fixture.task_start_mock.stop()
        mock_start = self.periodic_task_fixture.task_start_mock.start()
        periodic_processor = self._create_periodic_processor()
        periodic_processor.start()
        periodic_processor._maintenance_task = mock.MagicMock()

        mock_start.assert_called_once()

    @mock.patch.object(periodic_task.PeriodicTask, 'execute_ops',
                       new=mock.Mock())
    def test_reset_called_on_sighup(self):
        pid, c2p_read = self._create_periodic_processor_ipc_fork('reset')

        self.assert_process_running(pid)

        os.kill(pid, signal.SIGHUP)

        self.assert_ipc_mock_called(c2p_read)

    @mock.patch.object(periodic_task.PeriodicTask, 'execute_ops')
    def test_reset_fires_maintenance_task(self, execute_mock):
        periodic_processor = self._create_periodic_processor()

        periodic_processor._start_maintenance_task()
        execute_mock.reset_mock()

        periodic_processor.reset()

        execute_mock.assert_has_calls([mock.call(forced=True)])

    def test_reset_succeeeds_when_maintenance_task_not_setup(self):
        periodic_processor = self._create_periodic_processor()

        # NOTE(mpeterson): This tests that if calling reset without setting up
        # the maintenance task then it would not raise an exception and just
        # proceed as usual.
        periodic_processor.reset()

    @mock.patch.object(periodic_task.PeriodicTask, 'execute_ops')
    def test_start_fires_maintenance_task(self, execute_mock):
        periodic_processor = self._create_periodic_processor()

        periodic_processor.start()

        execute_mock.called_once_with([mock.call(forced=True)])

    def test_creates_pidfile(self):
        periodic_processor = self._create_periodic_processor()
        periodic_processor._create_pidfile()

        pidfile = str(periodic_processor.pidfile)
        self.assertTrue(os.path.isfile(pidfile))

        with open(pidfile) as f:
            pid = int(f.readline())

        self.assertEqual(pid, os.getpid())

        # NOTE(mpeterson): to avoid showing an expected exception while
        # running the next assert
        with mock.patch('neutron.agent.linux.daemon.LOG', autospec=True):
            self.assertRaises(
                SystemExit,
                worker.JournalPeriodicProcessor()._create_pidfile
            )

    @mock.patch.object(worker.JournalPeriodicProcessor, '_create_pidfile')
    @mock.patch.object(worker.JournalPeriodicProcessor, '_delete_pidfile')
    def test_pidfile_handling_on_start_stop(self, mock_create, mock_delete):
        periodic_processor = self._create_periodic_processor()
        periodic_processor.start()
        periodic_processor.stop()

        mock_create.assert_called_once()
        mock_delete.assert_called_once()

    def test_deletes_pidfile(self):
        atexit_mock = self.journal_thread_fixture.remock_atexit()

        periodic_processor = self._create_periodic_processor()
        periodic_processor.start()

        pidfile = str(periodic_processor.pidfile)
        self.assertTrue(os.path.isfile(pidfile))

        periodic_processor._delete_pidfile()

        self.assertFalse(os.path.isfile(pidfile))

        atexit_mock.assert_called_once_with(periodic_processor._delete_pidfile)

    def test_atexit_delete_pidfile_registered_only_once(self):
        atexit_mock = self.journal_thread_fixture.remock_atexit()
        periodic_processor = self._create_periodic_processor()

        for _ in range(0, 2):
            periodic_processor.start()
            periodic_processor.stop()

        atexit_mock.assert_called_once()


class OpenDaylightJournalThreadTest(base_v2.OpenDaylightTestCase):
    def setUp(self):
        super(OpenDaylightJournalThreadTest, self).setUp()
        self.journal = journal.OpenDaylightJournalThread()
        self.addCleanup(self.cleanup)

    @staticmethod
    def cleanup():
        journal.MAKE_URL.clear()

    def test_json_data(self):
        object_type = 'testobject'
        data = 'testdata'
        row = models.OpenDaylightJournal(object_type=object_type,
                                         object_uuid=uuidutils.generate_uuid(),
                                         operation=odl_const.ODL_CREATE,
                                         data=data)

        self.assertEqual("%ss" % object_type, self.journal._json_data(row)[1])

    def test_json_data_customized_url(self):
        object_type = 'randomtestobject'
        data = 'testdata'
        journal.register_url_builder(object_type, lambda row: row.object_type)
        row = models.OpenDaylightJournal(object_type=object_type,
                                         object_uuid=uuidutils.generate_uuid(),
                                         operation=odl_const.ODL_CREATE,
                                         data=data)

        url_param = self.journal._json_data(row)
        self.assertEqual(object_type, url_param[1])

    def test_entry_reset_retries_exceptions(self):
        with mock.patch.object(db, 'update_db_row_state') as m:
            self._test_retry_exceptions(journal.entry_reset, m)

    @test_db.in_session
    @mock.patch.object(client.OpenDaylightRestClient, 'sendjson',
                       mock.Mock(side_effect=Exception))
    def test__sync_entry_update_state_by_retry_count_on_exception(self):
        entry = db.create_pending_row(self.db_context, *self.UPDATE_ROW)
        self.journal._max_retry_count = 1
        self.assertEqual(entry.retry_count, 0)
        self.journal._sync_entry(self.db_context, entry)
        self.assertEqual(entry.retry_count, 1)
        self.assertEqual(entry.state, odl_const.PENDING)
        self.journal._sync_entry(self.db_context, entry)
        self.assertEqual(entry.retry_count, 1)
        self.assertEqual(entry.state, odl_const.FAILED)

    def _test__sync_entry_logs(self, log_type):
        entry = db.create_pending_row(self.db_context, *self.UPDATE_ROW)
        logger = self.useFixture(fixtures.FakeLogger())

        self.journal._sync_entry(self.db_context, entry)

        self.assertIn(log_type, logger.output)

    def test__sync_entry_logs_processing(self):
        self._test__sync_entry_logs(journal.LOG_PROCESSING)

    def test__sync_entry_logs_completed(self):
        self._test__sync_entry_logs(journal.LOG_COMPLETED)

    @mock.patch.object(client.OpenDaylightRestClient, 'sendjson',
                       mock.Mock(side_effect=Exception))
    def test__sync_entry_logs_failed(self):
        self._test__sync_entry_logs(journal.LOG_ERROR_PROCESSING)

    @mock.patch.object(journal.OpenDaylightJournalThread,
                       'sync_pending_entries')
    def test_terminate_journal_thread_correctly(self, mock_journal):
        self.journal_thread_fixture.journal_thread_mock.stop()
        self.addCleanup(self.journal_thread_fixture.journal_thread_mock.start)

        journal_thread = journal.OpenDaylightJournalThread(start_thread=True)

        journal_thread.stop(5)
        self.assertTrue(not journal_thread._odl_sync_thread.is_alive())
        mock_journal.assert_called_once()

    @mock.patch.object(journal.OpenDaylightJournalThread,
                       'sync_pending_entries')
    def test_allow_multiple_starts_gracefully(self, mock_journal):
        self.journal_thread_fixture.journal_thread_mock.stop()
        self.addCleanup(self.journal_thread_fixture.journal_thread_mock.start)
        journal_thread = journal.OpenDaylightJournalThread(start_thread=False)
        self.addCleanup(journal_thread.stop)
        journal_thread.start()

        try:
            journal_thread.start()
        except RuntimeError:
            self.fail('OpenDaylightJournalThread started twice')


def _raise_DBReferenceError(*args, **kwargs):
    args = [mock.Mock(unsafe=True)] * 4
    e = exception.DBReferenceError(*args)
    raise e


class JournalTest(base_v2.OpenDaylightTestCase):
    @mock.patch.object(dependency_validations, 'calculate')
    @mock.patch.object(journal.db, 'create_pending_row',
                       side_effect=_raise_DBReferenceError)
    def test_record_triggers_retry_on_reference_error(self, mock_create_row,
                                                      mock_calculate):
        args = [mock.Mock(unsafe=True)] * 5
        self.assertRaises(exception.RetryRequest, journal.record, *args)

    def test_entry_complete_retries_exceptions(self):
        with mock.patch.object(db, 'update_db_row_state') as m:
            self._test_retry_exceptions(journal.entry_complete, m)

    @test_db.in_session
    def _test_entry_complete(self, retention, expected_length):
        self.cfg.config(completed_rows_retention=retention, group='ml2_odl')
        db.create_pending_row(self.db_context, *test_db.DbTestCase.UPDATE_ROW)
        entry = db.get_all_db_rows(self.db_context)[-1]
        journal.entry_complete(self.db_context, entry)
        rows = db.get_all_db_rows(self.db_context)
        self.assertEqual(expected_length, len(rows))
        self.assertTrue(
            all(row.state == odl_const.COMPLETED for row in rows))

    def test_entry_complete_no_retention(self):
        self._test_entry_complete(0, 0)

    def test_entry_complete_with_retention(self):
        self._test_entry_complete(1, 1)

    def test_entry_complete_with_indefinite_retention(self):
        self._test_entry_complete(-1, 1)

    @test_db.in_session
    def test_entry_complete_with_retention_deletes_dependencies(self):
        self.cfg.config(completed_rows_retention=1, group='ml2_odl')
        db.create_pending_row(self.db_context, *test_db.DbTestCase.UPDATE_ROW)
        entry = db.get_all_db_rows(self.db_context)[-1]
        db.create_pending_row(self.db_context, *test_db.DbTestCase.UPDATE_ROW,
                              depending_on=[entry])
        dependant = db.get_all_db_rows(self.db_context)[-1]
        journal.entry_complete(self.db_context, entry)
        rows = db.get_all_db_rows(self.db_context)
        self.assertIn(entry, rows)
        self.assertEqual([], entry.dependencies)
        self.assertEqual([], dependant.depending_on)

    def test_entry_reset_retries_exceptions(self):
        with mock.patch.object(db, 'update_db_row_state') as m:
            self._test_retry_exceptions(journal.entry_reset, m)

    @test_db.in_session
    def test_entry_reset(self):
        db.create_pending_row(self.db_context, *test_db.DbTestCase.UPDATE_ROW)
        db.create_pending_row(self.db_context, *test_db.DbTestCase.UPDATE_ROW)
        entry = db.get_all_db_rows(self.db_context)[-1]
        entry.state = odl_const.PROCESSING
        self.db_context.session.merge(entry)
        self.db_context.session.flush()
        entry = db.get_all_db_rows(self.db_context)[-1]
        self.assertEqual(entry.state, odl_const.PROCESSING)
        journal.entry_reset(self.db_context, entry)
        rows = db.get_all_db_rows(self.db_context)
        self.assertEqual(2, len(rows))
        self.assertTrue(all(row.state == odl_const.PENDING for row in rows))

    def test_entry_set_retry_count_retries_exceptions(self):
        with mock.patch.object(db, 'update_pending_db_row_retry') as m:
            self._test_retry_exceptions(
                journal.entry_update_state_by_retry_count, m)

    @test_db.in_session
    def test_entry_set_retry_count(self):
        db.create_pending_row(self.db_context, *test_db.DbTestCase.UPDATE_ROW)
        entry_baseline = db.get_all_db_rows(self.db_context)[-1]
        db.create_pending_row(self.db_context, *test_db.DbTestCase.UPDATE_ROW)
        entry_target = db.get_all_db_rows(self.db_context)[-1]
        self.assertEqual(entry_target.retry_count, 0)
        self.assertEqual(entry_target.retry_count, entry_baseline.retry_count)
        self.assertEqual(entry_target.state, entry_baseline.state)

        journal.entry_update_state_by_retry_count(
            self.db_context, entry_target, 1)
        self.assertEqual(entry_target.retry_count, 1)
        self.assertEqual(entry_target.state, odl_const.PENDING)

        journal.entry_update_state_by_retry_count(
            self.db_context, entry_target, 1)
        self.assertEqual(entry_target.retry_count, 1)
        self.assertEqual(entry_target.state, odl_const.FAILED)
        self.assertNotEqual(entry_target.state, entry_baseline.state)
        self.assertNotEqual(entry_target.retry_count,
                            entry_baseline.retry_count)

    def test_record_logs_recording(self):
        logger = self.useFixture(fixtures.FakeLogger())
        journal.record(self.db_context, *self.UPDATE_ROW)
        for arg in self.UPDATE_ROW[0:3]:
            self.assertIn(arg, logger.output)

    def test_record_logs_dependencies(self):
        entry = db.create_pending_row(self.db_context, *self.UPDATE_ROW)

        logger = self.useFixture(fixtures.FakeLogger(level=logging.DEBUG))
        journal.record(self.db_context, *self.UPDATE_ROW)
        self.assertIn(str(entry.seqnum), logger.output)



# Copyright (c) 2017 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.callbacks import registry
from neutron_lib import constants as n_const
from neutron_lib.plugins import directory
from oslo_log import log as logging

from neutron.plugins.ml2 import driver_context

from networking_odl.common import constants
from networking_odl.dhcp import odl_dhcp_driver_base as driver_base


LOG = logging.getLogger(__name__)


@registry.has_registry_receivers
class OdlDhcpDriver(driver_base.OdlDhcpDriverBase):

    @registry.receives(constants.ODL_SUBNET, [constants.BEFORE_COMPLETE])
    def handle_subnet_event(self, resource, event, trigger, payload):
        context = payload.context
        operation = payload.metadata['operation']
        row = payload.metadata['row']
        if (operation == constants.ODL_CREATE or
                operation == constants.ODL_UPDATE):
            try:
                subnet_ctxt = self._get_subnet_context(context,
                                                       row.data['network_id'],
                                                       row.data['id'])
                self.create_or_delete_dhcp_port(subnet_ctxt)
            except Exception as e:
                LOG.error("Error while processing %s subnet %s: %s", operation,
                          row.data['id'], e)

    # Note(lajoskatona): When subnet locking was introduced in neutron
    # (see [0]) that caused that subnet deletion wasn't triggered dhcp port
    # deletion any more.
    # To workaround this let's catch subnet after_delete event and delete port
    # if that is a dhcp port.
    # The new functionality can't be added to handle_subnet_event as here
    # subnet context is not available.
    # [0]: https://review.opendev.org/713045
    @registry.receives(constants.ODL_SUBNET, [constants.AFTER_DELETE])
    def handle_subnet_delete_event(self, resource, event, trigger,
                                   payload):
        subnet = payload.latest_state
        context = payload.context
        operation = payload.metadata.get('operation')

        if event == constants.AFTER_DELETE:
            try:
                plugin = directory.get_plugin()
                network_id = subnet['network_id']
                device_id = constants.ODL_DEVICE_ID_START + '-' + subnet['id']
                filters = {
                    'network_id': [network_id],
                    'device_id': [device_id],
                    'device_owner': [n_const.DEVICE_OWNER_DHCP]
                }
                LOG.debug("Retrieving ODL DHCP port for deleted subnet %s",
                          subnet['id'])
                ports = plugin.get_ports(context, filters=filters)
                if ports:
                    self._delete_port_if_dhcp(plugin, context, ports[0])
            except Exception as e:
                LOG.error("Error while deleting %s subnet %s: %s", operation,
                          subnet['id'], e)

    def _delete_port_if_dhcp(self, plugin, context, port):
        dev_owner = port['device_owner']
        dev_id = port['device_id'] or ''
        device_id_type = constants.ODL_DEVICE_ID_START
        if (dev_owner == n_const.DEVICE_OWNER_DHCP and
                dev_id.startswith(device_id_type)):
            self._delete_port(plugin, context, port['id'])

    @registry.receives(constants.ODL_PORT, [constants.BEFORE_COMPLETE])
    def handle_port_update_event(self, resource, event, trigger,
                                 payload):
        context = payload.context
        operation = payload.metadata['operation']
        row = payload.metadata['row']

        if operation == constants.ODL_UPDATE:
            try:
                self._delete_if_dhcp_port(context, row)
            except Exception as e:
                device_id = row.data['device_id']
                subnet_id = device_id[13:] if device_id else ''
                LOG.error("Error while processing %s port %s of subnet %s: %s",
                          operation, row.data['id'], subnet_id, e)

    def _get_subnet_context(self, context, network_id, subnet_id):
        plugin = directory.get_plugin()
        network = plugin.get_network(context, network_id)
        subnet = plugin.get_subnet(context, subnet_id)
        return driver_context.SubnetContext(plugin, context,
                                            subnet, network)

    def _delete_if_dhcp_port(self, context, row):

        device_owner = row.data['device_owner']
        device_id = row.data['device_id']
        fixed_ips = row.data['fixed_ips']
        device_id_type = constants.ODL_DEVICE_ID_START
        if (device_owner and device_owner == n_const.DEVICE_OWNER_DHCP and
                device_id and
                device_id.startswith(device_id_type) and not fixed_ips):
            plugin = directory.get_plugin()
            self._delete_port(plugin, context, row.data['id'])



# Copyright (c) 2016 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
from string import Template

from neutron_lib.api.definitions import portbindings
from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib import constants as nl_const
from neutron_lib import context
from neutron_lib.plugins import directory
from neutron_lib.plugins.ml2 import api
from neutron_lib import worker
from oslo_config import cfg
from oslo_log import log
from oslo_serialization import jsonutils
from requests import codes
from requests import exceptions

from neutron.db import provisioning_blocks

from networking_odl.common import client as odl_client
from networking_odl.common import odl_features
from networking_odl.common import utils
from networking_odl.common import websocket_client as odl_ws_client
from networking_odl.journal import periodic_task
from networking_odl.ml2 import port_binding

cfg.CONF.import_group('ml2_odl', 'networking_odl.common.config')
LOG = log.getLogger(__name__)


class PseudoAgentDBBindingTaskBase(object):
    def __init__(self, worker):
        super(PseudoAgentDBBindingTaskBase, self).__init__()
        self._worker = worker

        # extract host/port from ODL URL and append hostconf_uri path
        hostconf_uri = utils.get_odl_url(cfg.CONF.ml2_odl.odl_hostconf_uri)
        LOG.debug("ODLPORTBINDING hostconfigs URI: %s", hostconf_uri)

        # TODO(mzmalick): disable port-binding for ODL lightweight testing
        self.odl_rest_client = odl_client.OpenDaylightRestClient.create_client(
            url=hostconf_uri)

    def _rest_get_hostconfigs(self):
        try:
            response = self.odl_rest_client.get()
            response.raise_for_status()
            hostconfigs = response.json()['hostconfigs']['hostconfig']
        except exceptions.ConnectionError:
            LOG.error("Cannot connect to the OpenDaylight Controller",
                      exc_info=True)
            return None
        except exceptions.HTTPError as e:
            # restconf returns 404 on operation when there is no entry
            if e.response.status_code == codes.not_found:
                LOG.debug("Response code not_found (404)"
                          " treated as an empty list")
                return []

            LOG.warning("REST/GET odl hostconfig failed, ",
                        exc_info=True)
            return None
        except KeyError:
            LOG.error("got invalid hostconfigs", exc_info=True)
            return None
        except Exception:
            LOG.warning("REST/GET odl hostconfig failed, ",
                        exc_info=True)
            return None
        else:
            if LOG.isEnabledFor(logging.DEBUG):
                _hconfig_str = jsonutils.dumps(
                    response, sort_keys=True, indent=4, separators=(',', ': '))
                LOG.debug("ODLPORTBINDING hostconfigs:\n%s", _hconfig_str)

        return hostconfigs

    def _get_and_update_hostconfigs(self, context=None):
        LOG.info("REST/GET hostconfigs from ODL")

        hostconfigs = self._rest_get_hostconfigs()

        if not hostconfigs:
            LOG.warning("ODL hostconfigs REST/GET failed, "
                        "will retry on next poll")
            return  # retry on next poll

        self._worker.update_agents_db(hostconfigs=hostconfigs)


@registry.has_registry_receivers
class PseudoAgentDBBindingPrePopulate(PseudoAgentDBBindingTaskBase):
    @registry.receives(resources.PORT,
                       [events.BEFORE_CREATE, events.BEFORE_UPDATE])
    def before_port_binding(self, resource, event, trigger, payload):
        LOG.debug("before_port resource %s event %s %s",
                  resource, event, payload)
        assert resource == resources.PORT
        assert event in [events.BEFORE_CREATE, events.BEFORE_UPDATE]
        ml2_plugin = trigger
        context = payload.context
        port = payload.latest_state

        host = nl_const.ATTR_NOT_SPECIFIED
        if port and portbindings.HOST_ID in port:
            host = port.get(portbindings.HOST_ID)
        if host == nl_const.ATTR_NOT_SPECIFIED or not host:
            return
        agent_type = PseudoAgentDBBindingWorker.L2_TYPE
        if self._worker.known_agent(host, agent_type):
            return
        agents = ml2_plugin.get_agents(
            context, filters={'agent_type': [agent_type], 'host': [host]})
        if agents and all(agent['alive'] for agent in agents):
            self._worker.add_known_agents(agents)
            LOG.debug("agents %s", agents)
            return

        # This host may not be created/updated by worker.
        # try to populate it.
        urlpath = "hostconfig/{0}/{1}".format(
            host, PseudoAgentDBBindingWorker.L2_TYPE)
        try:
            response = self.odl_rest_client.get(urlpath)
            response.raise_for_status()
        except Exception:
            LOG.warning("REST/GET odl hostconfig/%s failed.", host,
                        exc_info=True)
            return
        LOG.debug("response %s", response.json())
        hostconfig = response.json().get('hostconfig', [])
        if hostconfig:
            self._worker.update_agents_db_row(hostconfig[0])


class PseudoAgentDBBindingPeriodicTask(PseudoAgentDBBindingTaskBase):
    def __init__(self, worker):
        super(PseudoAgentDBBindingPeriodicTask, self).__init__(worker)

        # Start polling ODL restconf using maintenance thread.
        # default: 30s (should be <= agent keep-alive poll interval)
        self._periodic = periodic_task.PeriodicTask(
            'hostconfig', cfg.CONF.ml2_odl.restconf_poll_interval)
        self._periodic.register_operation(self._get_and_update_hostconfigs)
        self._periodic.start()


class PseudoAgentDBBindingWebSocket(PseudoAgentDBBindingTaskBase):
    def __init__(self, worker):
        super(PseudoAgentDBBindingWebSocket, self).__init__(worker)

        # Update hostconfig once for the configurations already present
        self._get_and_update_hostconfigs()
        odl_url = utils.get_odl_url()
        self._start_websocket(odl_url)

    def _start_websocket(self, odl_url):
        # OpenDaylight path to recieve websocket notifications on
        neutron_hostconfigs_path = """/neutron:neutron/neutron:hostconfigs"""

        self.odl_websocket_client = (
            odl_ws_client.OpenDaylightWebsocketClient.odl_create_websocket(
                odl_url, neutron_hostconfigs_path,
                odl_ws_client.ODL_OPERATIONAL_DATASTORE,
                odl_ws_client.ODL_NOTIFICATION_SCOPE_SUBTREE,
                self._process_websocket_recv,
                self._process_websocket_reconnect
            ))
        if self.odl_websocket_client is None:
            LOG.error("Error starting websocket thread")

    def _process_websocket_recv(self, payload, reconnect):
        # Callback for websocket notification
        LOG.debug("Websocket notification for hostconfig update")
        for event in odl_ws_client.EventDataParser.get_item(payload):
            try:
                operation, path, data = event.get_fields()
                if operation == event.OPERATION_DELETE:
                    host_id = event.extract_field(path, "neutron:host-id")
                    host_type = event.extract_field(path, "neutron:host-type")
                    if not host_id or not host_type:
                        LOG.warning("Invalid delete notification")
                        continue
                    self._worker.delete_agents_db_row(
                        host_id.strip("'"), host_type.strip("'"))
                elif operation == event.OPERATION_CREATE:
                    if 'hostconfig' in data:
                        hostconfig = data['hostconfig']
                        self.update_agents_db_row(hostconfig)
            except KeyError:
                LOG.warning("Invalid JSON for websocket notification",
                            exc_info=True)
                continue

    # TODO(rsood): Mixing restconf and websocket can cause race conditions
    def _process_websocket_reconnect(self, status):
        if status == odl_ws_client.ODL_WEBSOCKET_CONNECTED:
            # Get hostconfig data using restconf
            LOG.debug("Websocket notification on reconnection")
            self._get_and_update_hostconfigs()


class PseudoAgentDBBindingWorker(worker.BaseWorker):
    """Neutron Worker to update agentdb based on ODL hostconfig."""

    AGENTDB_BINARY = 'neutron-odlagent-portbinding'
    L2_TYPE = "ODL L2"

    # TODO(mzmalick): binary, topic and resource_versions to be provided
    # by ODL, Pending ODL NB patches.
    _AGENTDB_ROW = {
        'binary': AGENTDB_BINARY,
        'host': '',
        'topic': nl_const.L2_AGENT_TOPIC,
        'configurations': {},
        'resource_versions': '',
        'agent_type': L2_TYPE,
        'start_flag': True}

    def __init__(self):
        LOG.info("PseudoAgentDBBindingWorker init")
        self._old_agents = set()
        self._known_agents = set()
        self.agents_db = None
        super(PseudoAgentDBBindingWorker, self).__init__()

    def start(self):
        LOG.info("PseudoAgentDBBindingWorker starting")
        super(PseudoAgentDBBindingWorker, self).start()
        self._start()

    def stop(self):
        pass

    def wait(self):
        pass

    def reset(self):
        pass

    def _start(self):
        """Initialization."""
        LOG.debug("Initializing ODL Port Binding Worker")
        if cfg.CONF.ml2_odl.enable_websocket_pseudo_agentdb:
            self._websocket = PseudoAgentDBBindingWebSocket(self)
        else:
            self._periodic_task = (PseudoAgentDBBindingPeriodicTask(self))

    def known_agent(self, host_id, agent_type):
        agent = (host_id, agent_type)
        return agent in self._known_agents or agent in self._old_agents

    def add_known_agents(self, agents):
        for agent in agents:
            self._known_agents.add((agent['host'], agent['agent_type']))

    def update_agents_db(self, hostconfigs):
        LOG.debug("ODLPORTBINDING Updating agents DB with ODL hostconfigs")

        self._old_agents = self._known_agents
        self._known_agents = set()
        for host_config in hostconfigs:
            self._update_agents_db_row(host_config)

    def update_agents_db_row(self, host_config):
        self._old_agents = self._known_agents
        self._update_agents_db_row(host_config)

    def _update_agents_db_row(self, host_config):
        if self.agents_db is None:
            self.agents_db = directory.get_plugin()

        # Update one row in agent db
        host_id = host_config['host-id']
        host_type = host_config['host-type']
        config = host_config['config']
        try:
            agentdb_row = self._AGENTDB_ROW.copy()
            agentdb_row['host'] = host_id
            agentdb_row['agent_type'] = host_type
            agentdb_row['configurations'] = jsonutils.loads(config)
            if (host_id, host_type) in self._old_agents:
                agentdb_row.pop('start_flag', None)
            self.agents_db.create_or_update_agent(
                context.get_admin_context(), agentdb_row)
            self._known_agents.add((host_id, host_type))
        except Exception:
            LOG.exception("Unable to update agentdb.")

    def delete_agents_db_row(self, host_id, host_type):
        """Delete agent row."""
        try:
            filters = {'agent_type': [host_type],
                       'host': [host_id]}
            agent = self.agents_db.get_agent_objects(
                context.get_admin_context(), filters=filters)
            if not agent:
                return

            LOG.debug("Deleting Agent with Agent id: %s", agent[0]['id'])
            self.agents_db.delete_agent(
                context.get_admin_context(), agent[0]['id'])
            self._known_agents.remove((host_id, host_type))
        except Exception:
            LOG.exception("Unable to delete from agentdb.")


@registry.has_registry_receivers
class PseudoAgentDBBindingController(port_binding.PortBindingController):
    """Switch agnostic Port binding controller for OpenDayLight."""

    def __init__(self):
        """Initialization."""
        LOG.debug("Initializing ODL Port Binding Controller")
        super(PseudoAgentDBBindingController, self).__init__()
        self._worker = PseudoAgentDBBindingWorker()

    @registry.receives(resources.PROCESS, [events.BEFORE_SPAWN])
    def _before_spawn(self, resource, event, trigger, payload=None):
        self._prepopulate = PseudoAgentDBBindingPrePopulate(self._worker)

    def get_workers(self):
        return [self._worker]

    def _substitute_hconfig_tmpl(self, port_context, hconfig):
        # TODO(mzmalick): Explore options for inlines string splicing of
        #                 port-id to 14 bytes as required by vhostuser types
        port_id = port_context.current['id']
        conf = hconfig.get('configurations')
        vnics = conf.get('supported_vnic_types')
        if vnics is None:
            return hconfig
        for vnic in vnics:
            if vnic.get('vif_type') == portbindings.VIF_TYPE_VHOST_USER:
                details = vnic.get('vif_details')
                if details is None:
                    continue
                port_prefix = details.get('port_prefix')
                port_prefix = port_prefix[:14]
                subs_ids = {
                    # $IDENTIFER string substitution in hostconfigs JSON string
                    'PORT_ID': port_id[:(14 - len(port_prefix))],
                }
                # Substitute identifiers and Convert JSON string to dict
                hconfig_conf_json = Template(jsonutils.dumps(details))
                # pylint: disable=too-many-function-args
                substituted_str = hconfig_conf_json.safe_substitute(subs_ids)
                vnic['vif_details'] = jsonutils.loads(substituted_str)
        return hconfig

    def bind_port(self, port_context):
        """bind port using ODL host configuration."""
        # Get all ODL hostconfigs for this host and type
        agentdb = port_context.host_agents(PseudoAgentDBBindingWorker.L2_TYPE)

        if not agentdb:
            LOG.warning("No valid hostconfigs in agentsdb for host %s",
                        port_context.host)
            return

        for raw_hconfig in agentdb:
            # do any $identifier substitution
            hconfig = self._substitute_hconfig_tmpl(port_context, raw_hconfig)

            # Found ODL hostconfig for this host in agentdb
            LOG.debug("ODLPORTBINDING bind port with hostconfig: %s", hconfig)

            if self._hconfig_bind_port(port_context, hconfig):
                break  # Port binding suceeded!
            # Port binding failed!
            LOG.warning(
                "Failed to bind Port %(pid)s devid %(device_id)s "
                "owner %(owner)s for host %(host)s "
                "on network %(network)s.", {
                    'pid': port_context.current['id'],
                    'device_id': port_context.current['device_id'],
                    'owner': port_context.current['device_owner'],
                    'host': port_context.host,
                    'network': port_context.network.current['id']})
        else:  # No hostconfig found for host in agentdb.
            LOG.warning("No ODL hostconfigs for host %s found in agentdb",
                        port_context.host)

    def _hconfig_bind_port(self, port_context, hconfig):
        """bind port after validating odl host configuration."""
        valid_segment = None

        for segment in port_context.segments_to_bind:
            if self._is_valid_segment(segment, hconfig['configurations']):
                valid_segment = segment
                break
        else:
            LOG.debug("No valid segments found!")
            return False

        confs = hconfig['configurations']['supported_vnic_types']

        # nova provides vnic_type in port_context to neutron.
        # neutron provides supported vif_type for binding based on vnic_type
        # in this case ODL hostconfigs has the vif_type to bind for vnic_type
        vnic_type = port_context.current.get(portbindings.VNIC_TYPE)

        vif_details = None
        for conf in confs:
            if conf["vnic_type"] == vnic_type:
                vif_type = conf.get('vif_type', portbindings.VIF_TYPE_OVS)
                LOG.debug("Binding vnic:'%s' to vif:'%s'", vnic_type, vif_type)
                vif_details = conf.get('vif_details', {})
                break
        else:
            LOG.error(
                "Binding failed: unsupported VNIC %(vnic_type)s on %(host)s",
                {'vnic_type': vnic_type, 'host': port_context.host})
            return False

        if not vif_details:  # empty vif_details could be trouble, warn.
            LOG.warning("hostconfig:vif_details was empty!")

        LOG.debug("Bind port %(port)s on network %(network)s with valid "
                  "segment %(segment)s and VIF type %(vif_type)r "
                  "VIF details %(vif_details)r.",
                  {'port': port_context.current['id'],
                   'network': port_context.network.current['id'],
                   'segment': valid_segment, 'vif_type': vif_type,
                   'vif_details': vif_details})

        port_status = self._prepare_initial_port_status(port_context)
        port_context.set_binding(valid_segment[api.ID], vif_type,
                                 vif_details, status=port_status)

        return True

    def _prepare_initial_port_status(self, port_context):
        port_status = nl_const.PORT_STATUS_ACTIVE
        if odl_features.has(odl_features.OPERATIONAL_PORT_STATUS):
            port_status = nl_const.PORT_STATUS_DOWN
            provisioning_blocks.add_provisioning_component(
                port_context._plugin_context, port_context.current['id'],
                resources.PORT, provisioning_blocks.L2_AGENT_ENTITY)
        return port_status

    def _is_valid_segment(self, segment, conf):
        """Verify a segment is supported by ODL."""
        network_type = segment[api.NETWORK_TYPE]
        return network_type in conf['allowed_network_types']



# Copyright 2017, NEC Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Drop created_at column

Revision ID: 7cbef5a56298
Revises: eccd865b7d3a
Create Date: 2017-08-16 05:49:53.964988

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = '7cbef5a56298'
down_revision = 'eccd865b7d3a'


def upgrade():
    op.drop_column('opendaylightjournal', 'created_at')



# Copyright (c) 2017 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import helpers as log_helpers


def _build_func(client_method):
    @log_helpers.log_method_call
    def f(self, *args, **kwargs):
        self.journal.set_sync_event()

    f.__name__ = client_method
    return f


def _unboundmethod(func, cls):
    func.__qualname__ = cls.__qualname__ + '.' + func.__name__  # PEP 3155
    return func


def _get_method_name(op, resource):
    return op + '_' + resource + '_postcommit'


def _build_method(cls, resource):
    # add methods like the following:
    #
    #    @log_helpers.log_method_call
    #    def <method>_<resource>_postcommit(self, *args, **kwargs):
    #        self.journal.set_sync_event()

    operations = ['create', 'update', 'delete']
    for op in operations:
        client_method = _get_method_name(op, resource)
        if hasattr(cls, client_method) and client_method not in cls.__dict__:
            f = _build_func(client_method)
            unbound = _unboundmethod(f, cls)
            setattr(cls, client_method, unbound)


def _build_methods(cls, *resources):
    for resource in resources:
        _build_method(cls, resource)


def add_postcommit(*args):
    def postcommit(cls):
        _build_methods(cls, *args)
        return cls

    return postcommit



# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections

from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib.db import api as db_api
from oslo_log import log as logging
from oslo_utils import excutils

from networking_odl.common import constants as odl_const

LOG = logging.getLogger(__name__)

ODLResource = collections.namedtuple('ODLResource', ('singular', 'plural'))
_RESOURCE_MAPPING = {
    resources.SECURITY_GROUP: ODLResource(odl_const.ODL_SG, odl_const.ODL_SGS),
    resources.SECURITY_GROUP_RULE: ODLResource(odl_const.ODL_SG_RULE,
                                               odl_const.ODL_SG_RULES),
}
_OPERATION_MAPPING = {
    events.PRECOMMIT_CREATE: odl_const.ODL_CREATE,
    events.PRECOMMIT_UPDATE: odl_const.ODL_UPDATE,
    events.PRECOMMIT_DELETE: odl_const.ODL_DELETE,
    events.AFTER_CREATE: odl_const.ODL_CREATE,
    events.AFTER_UPDATE: odl_const.ODL_UPDATE,
    events.AFTER_DELETE: odl_const.ODL_DELETE,
}

LOG_TEMPLATE = ("(%(msg)s) with ODL_OPS (%(op)s) ODL_RES_TYPE (%(res_type)s) "
                "ODL_RES_ID (%(res_id)s)) ODL_RES_DICT (%(res_dict)s) "
                "DATA (%(data)s)")


def _log_on_callback(lvl, msg, op, res_type, res_id, res_dict, data):
    LOG.log(lvl, LOG_TEMPLATE,
            {'msg': msg, 'op': op, 'res_type': res_type, 'res_id': res_id,
             'res_dict': res_dict, 'data': data,
             'exc_info': lvl >= logging.ERROR})


class OdlSecurityGroupsHandler(object):

    def __init__(self, precommit, postcommit):
        assert postcommit is not None
        self._precommit = precommit
        self._postcommit = postcommit
        self._subscribe()

    def _subscribe(self):
        if self._precommit is not None:
            for event in (events.PRECOMMIT_CREATE, events.PRECOMMIT_DELETE):
                registry.subscribe(self.sg_callback_precommit,
                                   resources.SECURITY_GROUP, event)
                registry.subscribe(self.sg_callback_precommit,
                                   resources.SECURITY_GROUP_RULE, event)
            registry.subscribe(
                self.sg_callback_precommit, resources.SECURITY_GROUP,
                events.PRECOMMIT_UPDATE)

        for event in (events.AFTER_CREATE, events.AFTER_DELETE):
            registry.subscribe(self.sg_callback_postcommit,
                               resources.SECURITY_GROUP, event)
            registry.subscribe(self.sg_callback_postcommit,
                               resources.SECURITY_GROUP_RULE, event)

        registry.subscribe(self.sg_callback_postcommit,
                           resources.SECURITY_GROUP, events.AFTER_UPDATE)

    def _sg_callback(self, callback, resource, event, trigger, **kwargs):
        if 'payload' in kwargs:
            # TODO(boden): remove shim once all callbacks use payloads
            context = kwargs['payload'].context
            res = kwargs['payload'].latest_state
            res_id = kwargs['payload'].resource_id
            copy_kwargs = kwargs
        else:
            context = kwargs['context']
            res = kwargs.get(resource)
            res_id = kwargs.get("%s_id" % resource)
            copy_kwargs = kwargs.copy()
            copy_kwargs.pop('context')

        if res_id is None:
            res_id = res.get('id')
        odl_res_type = _RESOURCE_MAPPING[resource]

        odl_ops = _OPERATION_MAPPING[event]
        odl_res_dict = None if res is None else {odl_res_type.singular: res}

        _log_on_callback(logging.DEBUG, "Calling callback", odl_ops,
                         odl_res_type, res_id, odl_res_dict,
                         {'security_group_rule': res,
                          'security_group_rule_id': res_id})
        try:
            callback(context, odl_ops, odl_res_type, res_id, odl_res_dict,
                     **copy_kwargs)
        except Exception as e:
            # In case of precommit, neutron registry notification caller
            # doesn't log its exception. In networking-odl case, we don't
            # normally throw exception. So log it here for debug
            with excutils.save_and_reraise_exception():
                if not db_api.is_retriable(e):
                    _log_on_callback(
                        logging.ERROR, "Exception from callback",
                        odl_ops, odl_res_type, res_id, odl_res_dict,
                        {'security_group_rule': res,
                         'security_group_rule_id': res_id})

    def sg_callback_precommit(self, resource, event, trigger, **kwargs):
        self._sg_callback(self._precommit, resource, event, trigger, **kwargs)

    def sg_callback_postcommit(self, resource, event, trigger, **kwargs):
        self._sg_callback(self._postcommit, resource, event, trigger, **kwargs)



# Copyright (c) 2013-2014 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import multiprovidernet as mpnet_apidef
from neutron_lib.api.definitions import provider_net as providernet
from neutron_lib import constants as p_const
from neutron_lib.plugins import constants as nlib_const
from neutron_lib.plugins.ml2 import api
from oslo_config import cfg
from oslo_log import helpers as log_helpers
from oslo_log import log as logging

from networking_odl.common import callback
from networking_odl.common import config as odl_conf
from networking_odl.common import constants as odl_const
from networking_odl.common import odl_features
from networking_odl.common import postcommit
from networking_odl.dhcp import odl_dhcp_driver as dhcp_driver
from networking_odl.journal import base_driver
from networking_odl.journal import full_sync
from networking_odl.journal import journal
from networking_odl.journal import worker
from networking_odl.ml2 import port_binding
from networking_odl.ml2 import port_status_update
from networking_odl.qos import qos_driver_v2 as qos_driver
from networking_odl.trunk import trunk_driver_v2 as trunk_driver

LOG = logging.getLogger(__name__)


@postcommit.add_postcommit('network', 'subnet', 'port')
class OpenDaylightMechanismDriver(api.MechanismDriver,
                                  base_driver.ResourceBaseDriver):
    """OpenDaylight Python Driver for Neutron.

    This code is the backend implementation for the OpenDaylight ML2
    MechanismDriver for OpenStack Neutron.
    """
    RESOURCES = {
        odl_const.ODL_SG: odl_const.ODL_SGS,
        odl_const.ODL_SG_RULE: odl_const.ODL_SG_RULES,
        odl_const.ODL_NETWORK: odl_const.ODL_NETWORKS,
        odl_const.ODL_SUBNET: odl_const.ODL_SUBNETS,
        odl_const.ODL_PORT: odl_const.ODL_PORTS
    }
    plugin_type = nlib_const.CORE

    def initialize(self):
        LOG.debug("Initializing OpenDaylight ML2 driver")
        cfg.CONF.register_opts(odl_conf.odl_opts, "ml2_odl")
        self.sg_handler = callback.OdlSecurityGroupsHandler(
            self.sync_from_callback_precommit,
            self.sync_from_callback_postcommit)
        self.journal = journal.OpenDaylightJournalThread()
        self.port_binding_controller = port_binding.PortBindingManager.create()
        self.trunk_driver = trunk_driver.OpenDaylightTrunkDriverV2.create()
        if cfg.CONF.ml2_odl.enable_dhcp_service:
            self.dhcp_driver = dhcp_driver.OdlDhcpDriver()

        full_sync.register(nlib_const.CORE, self.RESOURCES)
        odl_features.init()

        if odl_const.ODL_QOS in cfg.CONF.ml2.extension_drivers:
            qos_driver.OpenDaylightQosDriver.create()

    def get_workers(self):
        workers = [port_status_update.OdlPortStatusUpdate(),
                   worker.JournalPeriodicProcessor()]
        workers += self.port_binding_controller.get_workers()
        return workers

    @staticmethod
    def _record_in_journal(context, object_type, operation, data=None):
        if data is None:
            data = context.current
        journal.record(context._plugin_context, object_type,
                       context.current['id'], operation, data,
                       ml2_context=context)

    @log_helpers.log_method_call
    def create_network_precommit(self, context):
        OpenDaylightMechanismDriver._record_in_journal(
            context, odl_const.ODL_NETWORK, odl_const.ODL_CREATE)

    @log_helpers.log_method_call
    def create_subnet_precommit(self, context):
        OpenDaylightMechanismDriver._record_in_journal(
            context, odl_const.ODL_SUBNET, odl_const.ODL_CREATE)

    @log_helpers.log_method_call
    def create_port_precommit(self, context):
        OpenDaylightMechanismDriver._record_in_journal(
            context, odl_const.ODL_PORT, odl_const.ODL_CREATE)

    @log_helpers.log_method_call
    def update_network_precommit(self, context):
        OpenDaylightMechanismDriver._record_in_journal(
            context, odl_const.ODL_NETWORK, odl_const.ODL_UPDATE)

    @log_helpers.log_method_call
    def update_subnet_precommit(self, context):
        OpenDaylightMechanismDriver._record_in_journal(
            context, odl_const.ODL_SUBNET, odl_const.ODL_UPDATE)

    @log_helpers.log_method_call
    def update_port_precommit(self, context):
        OpenDaylightMechanismDriver._record_in_journal(
            context, odl_const.ODL_PORT, odl_const.ODL_UPDATE)

    @log_helpers.log_method_call
    def delete_network_precommit(self, context):
        OpenDaylightMechanismDriver._record_in_journal(
            context, odl_const.ODL_NETWORK, odl_const.ODL_DELETE, data=[])

    @log_helpers.log_method_call
    def delete_subnet_precommit(self, context):
        # Use the journal row's data field to store parent object
        # uuids. This information is required for validation checking
        # when deleting parent objects.
        new_context = [context.current['network_id']]
        OpenDaylightMechanismDriver._record_in_journal(
            context, odl_const.ODL_SUBNET, odl_const.ODL_DELETE,
            data=new_context)

    @log_helpers.log_method_call
    def delete_port_precommit(self, context):
        # Use the journal row's data field to store parent object
        # uuids. This information is required for validation checking
        # when deleting parent objects.
        new_context = [context.current['network_id']]
        for subnet in context.current['fixed_ips']:
            new_context.append(subnet['subnet_id'])
        OpenDaylightMechanismDriver._record_in_journal(
            context, odl_const.ODL_PORT, odl_const.ODL_DELETE,
            data=new_context)

    def _sync_security_group_create_precommit(
            self, context, operation, object_type, res_id, sg_dict):
        journal.record(context, object_type, sg_dict['id'], operation, sg_dict)

        # NOTE(yamahata): when security group is created, default rules
        # are also created.
        for rule in sg_dict['security_group_rules']:
            journal.record(context, odl_const.ODL_SG_RULE, rule['id'],
                           odl_const.ODL_CREATE, rule)

    @log_helpers.log_method_call
    def sync_from_callback_precommit(self, context, operation, res_type,
                                     res_id, resource_dict, **kwargs):
        # TODO(lajoskatona): remove this when everything runs with payload
        if 'payload' in kwargs:
            ids = kwargs['payload'].metadata
            context = kwargs['payload'].context
        else:
            ids = kwargs
        object_type = res_type.singular
        if resource_dict is not None:
            resource_dict = resource_dict[object_type]

        if (operation == odl_const.ODL_CREATE and
                object_type == odl_const.ODL_SG):
            self._sync_security_group_create_precommit(
                context, operation, object_type, res_id, resource_dict)
            return

        object_uuid = (resource_dict.get('id')
                       if operation == 'create' else res_id)
        data = resource_dict

        if (operation == odl_const.ODL_DELETE):
            # NOTE(yamahata): DB auto deletion
            # Security Group Rule under this Security Group needs to
            # be deleted. At NeutronDB layer rules are auto deleted with
            # cascade='all,delete'.
            if (object_type == odl_const.ODL_SG):
                for rule_id in ids['security_group_rule_ids']:
                    journal.record(context, odl_const.ODL_SG_RULE,
                                   rule_id, odl_const.ODL_DELETE,
                                   [object_uuid])
            elif (object_type == odl_const.ODL_SG_RULE):
                # Set the parent security group id so that dependencies
                # to this security rule deletion can be properly found
                # in the journal.
                data = [ids['security_group_id']]

        assert object_uuid is not None
        journal.record(context, object_type, object_uuid,
                       operation, data)

    def sync_from_callback_postcommit(self, context, operation, res_type,
                                      res_id, resource_dict, payload):
        self._postcommit(context)

    def _postcommit(self, context):
        self.journal.set_sync_event()

    @log_helpers.log_method_call
    def bind_port(self, context):
        """Set binding for a valid segments

        """
        return self.port_binding_controller.bind_port(port_context=context)

    def check_vlan_transparency(self, context):
        """Check VLAN transparency

        """
        # TODO(yamahata): This should be odl service provider dependent
        # introduce ODL yang model for ODL to report which network types
        # are vlan-transparent.
        # VLAN and FLAT cases, we don't know if the underlying network
        # supports QinQ or VLAN.
        # For now, netvirt supports only vxlan tunneling.
        VLAN_TRANSPARENT_NETWORK_TYPES = [p_const.TYPE_VXLAN]
        network = context.current
        # see TypeManager._extend_network_dict_provider()
        # single providernet
        if providernet.NETWORK_TYPE in network:
            return (network[providernet.NETWORK_TYPE] in
                    VLAN_TRANSPARENT_NETWORK_TYPES)
        # multi providernet
        segments = network.get(mpnet_apidef.SEGMENTS)
        if segments is None:
            return True
        return all(segment[providernet.NETWORK_TYPE]
                   in VLAN_TRANSPARENT_NETWORK_TYPES
                   for segment in segments)



# Copyright (c) 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import datetime

from neutron_lib.db import api as db_api
from oslo_log import log as logging
from sqlalchemy import asc
from sqlalchemy import bindparam
from sqlalchemy.ext import baked
from sqlalchemy import func
from sqlalchemy import or_
from sqlalchemy.orm import aliased


from networking_odl.common import constants as odl_const
from networking_odl.db import models

LOG = logging.getLogger(__name__)

bakery = baked.bakery()


def get_pending_or_processing_ops(context, object_uuid, operation=None):
    # NOTE (sai): For performance reasons, we expect this method to use baked
    # query (http://docs.sqlalchemy.org/en/latest/orm/extensions/baked.html)
    baked_query = bakery(lambda s: s.query(
        models.OpenDaylightJournal))
    baked_query += lambda q: q.filter(
        or_(models.OpenDaylightJournal.state == odl_const.PENDING,
            models.OpenDaylightJournal.state == odl_const.PROCESSING),
        models.OpenDaylightJournal.object_uuid == bindparam('uuid'))
    if operation:
        if isinstance(operation, (list, tuple)):
            baked_query += lambda q: q.filter(
                models.OpenDaylightJournal.operation.in_(bindparam('op',
                                                         expanding=True)))
        else:
            baked_query += lambda q: q.filter(
                models.OpenDaylightJournal.operation == bindparam('op'))

    return baked_query(context.session).params(
        uuid=object_uuid, op=operation).all()


def get_pending_delete_ops_with_parent(context, object_type, parent_id):
    rows = context.session.query(models.OpenDaylightJournal).filter(
        or_(models.OpenDaylightJournal.state == odl_const.PENDING,
            models.OpenDaylightJournal.state == odl_const.PROCESSING),
        models.OpenDaylightJournal.object_type == object_type,
        models.OpenDaylightJournal.operation == odl_const.ODL_DELETE
    ).all()

    return (row for row in rows if parent_id in row.data)


def get_all_db_rows(context):
    return context.session.query(models.OpenDaylightJournal).all()


@db_api.CONTEXT_READER
def get_all_db_rows_by_state(context, state):
    return context.session.query(models.OpenDaylightJournal).filter_by(
        state=state).all()


# Retry deadlock exception for Galera DB.
# If two (or more) different threads call this method at the same time, they
# might both succeed in changing the same row to pending, but at least one
# of them will get a deadlock from Galera and will have to retry the operation.
@db_api.retry_if_session_inactive()
@db_api.CONTEXT_WRITER.savepoint
def get_oldest_pending_db_row_with_lock(context):
    # NOTE (sai): For performance reasons, we expect this method to use baked
    # query (http://docs.sqlalchemy.org/en/latest/orm/extensions/baked.html)
    journal_dep = aliased(models.OpenDaylightJournal)
    dep_query = bakery(lambda s1: s1.query(journal_dep))
    dep_query += lambda q: q.filter(
        models.OpenDaylightJournal.seqnum == journal_dep.seqnum)
    dep_query += lambda q: q.outerjoin(
        journal_dep.depending_on, aliased=True)
    dep_query += lambda q: q.filter(
        or_(models.OpenDaylightJournal.state == odl_const.PENDING,
            models.OpenDaylightJournal.state == odl_const.PROCESSING))
    row = bakery(lambda s2: s2.query(models.OpenDaylightJournal))
    row += lambda q: q.filter(
        models.OpenDaylightJournal.state == odl_const.PENDING,
        ~ (dep_query._as_query(q.session)).exists())
    row += lambda q: q.order_by(
        asc(models.OpenDaylightJournal.last_retried))
    row = row(context.session).first()
    if row:
        update_db_row_state(context, row, odl_const.PROCESSING)

    return row


def delete_dependency(context, entry):
    """Delete dependency upon the given ID"""
    conn = context.session.connection()
    stmt = models.journal_dependencies.delete(
        models.journal_dependencies.c.depends_on == entry.seqnum)
    conn.execute(stmt)
    context.session.expire_all()


def update_db_row_state(context, row, state, flush=True):
    row.state = state
    context.session.merge(row)
    if flush:
        context.session.flush()


def update_pending_db_row_retry(context, row, retry_count):
    if row.retry_count >= retry_count:
        update_db_row_state(context, row, odl_const.FAILED)
    else:
        row.retry_count += 1
        update_db_row_state(context, row, odl_const.PENDING)


def delete_row(context, row=None, row_id=None, flush=True):
    if row_id:
        row = context.session.query(models.OpenDaylightJournal).filter_by(
            seqnum=row_id).one()
    if row:
        context.session.delete(row)
        if flush:
            context.session.flush()


def create_pending_row(context, object_type, object_uuid,
                       operation, data, depending_on=None):
    if depending_on is None:
        depending_on = []
    row = models.OpenDaylightJournal(object_type=object_type,
                                     object_uuid=object_uuid,
                                     operation=operation, data=data,
                                     state=odl_const.PENDING,
                                     depending_on=depending_on)
    context.session.add(row)
    # Keep session flush for unit tests. NOOP for L2/L3 events since calls are
    # made inside database session transaction with subtransactions=True.
    context.session.flush()
    return row


@db_api.CONTEXT_WRITER.savepoint
def delete_pending_rows(context, operations_to_delete):
    context.session.query(models.OpenDaylightJournal).filter(
        models.OpenDaylightJournal.operation.in_(operations_to_delete),
        models.OpenDaylightJournal.state == odl_const.PENDING).delete(
        synchronize_session=False)
    context.session.expire_all()


def _update_periodic_task_state(context, expected_state, state, task):
    row = context.session.query(models.OpenDaylightPeriodicTask).filter_by(
        state=expected_state,
        task=task).with_for_update().one_or_none()

    if row is None:
        return False

    row.state = state
    return True


def was_periodic_task_executed_recently(context, task, interval):
    now = context.session.execute(func.now()).scalar()
    delta = datetime.timedelta(seconds=interval)
    row = context.session.query(models.OpenDaylightPeriodicTask).filter(
        models.OpenDaylightPeriodicTask.task == task,
        (now - delta >= (models.OpenDaylightPeriodicTask.lock_updated))
    ).one_or_none()

    return bool(row is None)


def lock_periodic_task(context, task):
    return _update_periodic_task_state(context, odl_const.PENDING,
                                       odl_const.PROCESSING, task)


def unlock_periodic_task(context, task):
    return _update_periodic_task_state(context, odl_const.PROCESSING,
                                       odl_const.PENDING, task)


def update_periodic_task(context, task, operation=None):
    """Update the current periodic task details.

    The function assumes the lock is held, so it mustn't be run outside of a
    locked context.
    """
    op_text = None
    if operation:
        op_text = operation.__name__

    row = context.session.query(models.OpenDaylightPeriodicTask).filter_by(
        task=task).one()
    row.processing_operation = op_text


@db_api.CONTEXT_WRITER.savepoint
def delete_rows_by_state_and_time(context, state, time_delta):
    # NOTE(mpeterson): The reason behind deleting one-by-one is that InnoDB
    # ignores the WHERE clause to issue a LOCK when executing a DELETE. By
    # executing each operation indepently, we minimize exposures to DEADLOCKS.
    now = context.session.execute(func.now()).scalar()
    rows = context.session.query(models.OpenDaylightJournal).filter(
        models.OpenDaylightJournal.state == state,
        models.OpenDaylightJournal.last_retried < now - time_delta).all()
    for row in rows:
        delete_row(context, row, flush=False)
    context.session.expire_all()


@db_api.CONTEXT_WRITER.savepoint
def reset_processing_rows(context, max_timedelta):
    # NOTE(mpeterson): The reason behind updating one-by-one is that InnoDB
    # ignores the WHERE clause to issue a LOCK when executing an UPDATE. By
    # executing each operation indepently, we minimize exposures to DEADLOCKS.
    now = context.session.execute(func.now()).scalar()
    max_timedelta = datetime.timedelta(seconds=max_timedelta)
    rows = context.session.query(models.OpenDaylightJournal).filter(
        models.OpenDaylightJournal.last_retried < now - max_timedelta,
        models.OpenDaylightJournal.state == odl_const.PROCESSING).all()
    for row in rows:
        update_db_row_state(context, row, odl_const.PENDING, flush=False)

    return len(rows)



# Copyright (c) 2017 Ericsson India Global Service Pvt Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from oslo_config import cfg
from oslo_log import helpers as log_helpers
from oslo_log import log as logging

from networking_l2gw.services.l2gateway.common import constants
from networking_l2gw.services.l2gateway import service_drivers
from networking_odl.common import constants as odl_const
from networking_odl.common import postcommit
from networking_odl.journal import full_sync
from networking_odl.journal import journal


cfg.CONF.import_group('ml2_odl', 'networking_odl.common.config')

LOG = logging.getLogger(__name__)

L2GW_RESOURCES = {
    odl_const.ODL_L2GATEWAY: odl_const.ODL_L2GATEWAYS,
    odl_const.ODL_L2GATEWAY_CONNECTION: odl_const.ODL_L2GATEWAY_CONNECTIONS
}


@postcommit.add_postcommit('l2_gateway', 'l2_gateway_connection')
class OpenDaylightL2gwDriver(service_drivers.L2gwDriver):
    """OpenDaylight L2Gateway Service Driver

    This code is the openstack driver for exciting the OpenDaylight L2GW
    facility.
    """

    def __init__(self, service_plugin, validator=None):
        super(OpenDaylightL2gwDriver, self).__init__(service_plugin, validator)
        self.service_plugin = service_plugin
        self.journal = journal.OpenDaylightJournalThread()
        full_sync.register(constants.L2GW, L2GW_RESOURCES)
        LOG.info("ODL: Started OpenDaylight L2Gateway V2 driver")

    @property
    def service_type(self):
        return constants.L2GW

    @log_helpers.log_method_call
    def create_l2_gateway_precommit(self, context, l2_gateway):
        journal.record(context, odl_const.ODL_L2GATEWAY,
                       l2_gateway['id'], odl_const.ODL_CREATE,
                       l2_gateway)

    @log_helpers.log_method_call
    def update_l2_gateway_precommit(self, context, l2_gateway):
        journal.record(context, odl_const.ODL_L2GATEWAY,
                       l2_gateway['id'], odl_const.ODL_UPDATE,
                       l2_gateway)

    @log_helpers.log_method_call
    def delete_l2_gateway_precommit(self, context, l2_gateway_id):
        journal.record(context, odl_const.ODL_L2GATEWAY,
                       l2_gateway_id, odl_const.ODL_DELETE,
                       l2_gateway_id)

    @log_helpers.log_method_call
    def create_l2_gateway_connection_precommit(self, context,
                                               l2_gateway_connection):
        odl_l2_gateway_connection = copy.deepcopy(l2_gateway_connection)
        odl_l2_gateway_connection['gateway_id'] = (
            l2_gateway_connection['l2_gateway_id'])
        odl_l2_gateway_connection.pop('l2_gateway_id')
        journal.record(context, odl_const.ODL_L2GATEWAY_CONNECTION,
                       odl_l2_gateway_connection['id'],
                       odl_const.ODL_CREATE,
                       odl_l2_gateway_connection)

    @log_helpers.log_method_call
    def delete_l2_gateway_connection_precommit(self, context,
                                               l2_gateway_connection_id):
        journal.record(context, odl_const.ODL_L2GATEWAY_CONNECTION,
                       l2_gateway_connection_id,
                       odl_const.ODL_DELETE,
                       l2_gateway_connection_id)



#
# Copyright (C) 2016 Red Hat, Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.
#

from neutron_lib.db import api as db_api
from neutron_lib.plugins import directory
import requests

from networking_odl.common import client
from networking_odl.common import constants as odl_const
from networking_odl.db import db
from networking_odl.journal import base_driver
from networking_odl.journal import journal

# Define which pending operation types should be deleted
_CANARY_NETWORK_ID = "bd8db3a8-2b30-4083-a8b3-b3fd46401142"
_CANARY_TENANT_ID = "bd8db3a8-2b30-4083-a8b3-b3fd46401142"
_CANARY_NETWORK_DATA = {'id': _CANARY_NETWORK_ID,
                        'tenant_id': _CANARY_TENANT_ID,
                        'name': 'Sync Canary Network',
                        'admin_state_up': False}
_OPS_TO_DELETE_ON_SYNC = (odl_const.ODL_CREATE, odl_const.ODL_UPDATE)
_CLIENT = client.OpenDaylightRestClientGlobal()

_ORDERED_ODL_RESOURCES = (
    odl_const.ODL_SG,
    odl_const.ODL_SG_RULE,
    odl_const.ODL_NETWORK,
    odl_const.ODL_SUBNET,
    odl_const.ODL_ROUTER,
    odl_const.ODL_PORT,
    odl_const.ODL_FLOATINGIP,
    odl_const.ODL_QOS_POLICY,
    odl_const.ODL_TRUNK,
    odl_const.ODL_BGPVPN,
    odl_const.ODL_BGPVPN_NETWORK_ASSOCIATION,
    odl_const.ODL_BGPVPN_ROUTER_ASSOCIATION,
    odl_const.ODL_SFC_FLOW_CLASSIFIER,
    odl_const.ODL_SFC_PORT_PAIR,
    odl_const.ODL_SFC_PORT_PAIR_GROUP,
    odl_const.ODL_SFC_PORT_CHAIN,
    odl_const.ODL_L2GATEWAY,
    odl_const.ODL_L2GATEWAY_CONNECTION,
)


# TODO(rajivk): Remove this variable, while fixing recovery
ALL_RESOURCES = {}

FULL_SYNC_RESOURCES = {}


def register(driver, resources, handler=None):
    def default_handler(context, resource_type):
        return get_resources(context, driver, resources[resource_type])

    ALL_RESOURCES[driver] = resources
    handler = handler or default_handler
    for resource in resources:
        FULL_SYNC_RESOURCES[resource] = handler


@db_api.retry_if_session_inactive()
@db_api.CONTEXT_WRITER.savepoint
def full_sync(context):
    if not _full_sync_needed(context):
        return

    db.delete_pending_rows(context, _OPS_TO_DELETE_ON_SYNC)

    for resource_type in _ORDERED_ODL_RESOURCES:
        handler = FULL_SYNC_RESOURCES.get(resource_type)
        if handler:
            _sync_resources(context, resource_type, handler)

    journal.record(context, odl_const.ODL_NETWORK, _CANARY_NETWORK_ID,
                   odl_const.ODL_CREATE, _CANARY_NETWORK_DATA)


def _full_sync_needed(context):
    return (_canary_network_missing_on_odl() and
            _canary_network_not_in_journal(context))


def _canary_network_missing_on_odl():
    # Try to reach the ODL server, sometimes it might be up & responding to
    # HTTP calls but inoperative..
    client = _CLIENT.get_client()
    response = client.get(odl_const.ODL_NETWORKS)
    response.raise_for_status()

    response = client.get(odl_const.ODL_NETWORKS + "/" + _CANARY_NETWORK_ID)
    if response.status_code == requests.codes.not_found:
        return True

    # In case there was an error raise it up because we don't know how to deal
    # with it..
    response.raise_for_status()
    return False


def _canary_network_not_in_journal(context):
    return not db.get_pending_or_processing_ops(
        context, _CANARY_NETWORK_ID, operation=odl_const.ODL_CREATE)


def get_resources_require_id(plugin, context, get_resources_for_id,
                             method_name_for_resource):

    dep_id_resources = get_resources_for_id(context)
    resources = []
    for d_resource in dep_id_resources:
        obj_getter = getattr(plugin, method_name_for_resource)
        resource = obj_getter(context, d_resource['id'])
        if resource:
            resources.extend(resource)

    return resources


def get_resources(context, plugin_type, resource_type):
    plugin = directory.get_plugin(plugin_type)
    obj_getter = getattr(plugin, 'get_%s' % resource_type)
    return obj_getter(context)


def _sync_resources(context, object_type, handler):
    resources = handler(context, object_type)
    for resource in resources:
        journal.record(context, object_type, resource['id'],
                       odl_const.ODL_CREATE, resource)


@db_api.retry_if_session_inactive()
# TODO(rajivk): Change name from sync_resource to _sync_resources
# once, we are completely moved to new sync mechanism to plug new syncing
# mechanism.
def sync_resources(context, resource_type):
    driver = base_driver.get_driver(resource_type)
    resources = driver.get_resources_for_full_sync(context, resource_type)
    with db_api.CONTEXT_WRITER.savepoint.using(context):
        for resource in resources:
            journal.record(context, resource_type, resource['id'],
                           odl_const.ODL_CREATE, resource)



# Copyright (c) 2015 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""OpenDaylight Neutron mechanism driver refactor

Revision ID: 37e242787ae5
Revises: 247501328046
Create Date: 2015-10-30 22:09:27.221767

"""
from neutron.db import migration

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '37e242787ae5'
down_revision = '247501328046'

# milestone identifier, used by neutron-db-manage
neutron_milestone = [migration.MITAKA]


def upgrade():
    op.create_table(
        'opendaylightjournal',
        sa.Column('id', sa.String(36), primary_key=True),
        sa.Column('object_type', sa.String(36), nullable=False),
        sa.Column('object_uuid', sa.String(36), nullable=False),
        sa.Column('operation', sa.String(36), nullable=False),
        sa.Column('data', sa.PickleType, nullable=True),
        sa.Column('state',
                  sa.Enum('pending', 'processing', 'failed', 'completed',
                          name='state'),
                  nullable=False, default='pending'),
        sa.Column('retry_count', sa.Integer, default=0),
        sa.Column('created_at', sa.DateTime, default=sa.func.now()),
        sa.Column('last_retried', sa.TIMESTAMP, server_default=sa.func.now(),
                  onupdate=sa.func.now())
    )



# Copyright (c) 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from networking_odl._i18n import _
from networking_odl.common import constants as odl_const
from networking_odl.db import db


def _get_delete_dependencies(context, object_type, object_uuid):
    """Get dependent operations for a delete operation.

    Return any operations that pertain to the delete: Either create
    or update operations on the same object, or delete operations on other
    objects that depend on the deleted object.
    """
    # Get any pending or processing create or update ops on the row itself
    deps = db.get_pending_or_processing_ops(
        context, object_uuid, operation=(odl_const.ODL_UPDATE,
                                         odl_const.ODL_CREATE))

    # Get dependent operations of other dependent types
    dependent_resource_types = _DELETE_DEPENDENCIES.get(object_type)
    if dependent_resource_types is not None:
        for resource_type in dependent_resource_types:
            deps.extend(db.get_pending_delete_ops_with_parent(
                context, resource_type, object_uuid))

    return deps


def _get_older_operations(context, object_ids):
    """Get any older operations.

    Return any operations still in the queue for the given ID(s).
    """
    if not isinstance(object_ids, (list, tuple)):
        object_ids = (object_ids,)

    deps = []
    for object_id in object_ids:
        deps.extend(
            db.get_pending_or_processing_ops(context, object_id))

    return deps


def _generate_subnet_deps(data):
    return data['network_id']


def _generate_port_deps(data):
    object_ids = set(fixed_ip['subnet_id'] for fixed_ip in data['fixed_ips'])
    object_ids = list(object_ids)
    object_ids.append(data['network_id'])
    qos_policy_id = data.get('qos_policy_id')
    if qos_policy_id is not None:
        object_ids.append(qos_policy_id)
    return object_ids


def _generate_network_deps(data):
    return data.get('qos_policy_id')


def _generate_sg_rule_deps(data):
    return data['security_group_id']


def _generate_router_deps(data):
    return data['gw_port_id']


def _generate_floatingip_deps(data):
    object_ids = []
    network_id = data.get('floating_network_id')
    if network_id is not None:
        object_ids.append(network_id)

    port_id = data.get('port_id')
    if port_id is not None:
        object_ids.append(port_id)

    router_id = data.get('router_id')
    if router_id is not None:
        object_ids.append(router_id)

    return object_ids


def _generate_trunk_deps(data):
    portids = [subport['port_id'] for subport in data['sub_ports']]
    portids.append(data['port_id'])
    return portids


def _generate_l2gateway_connection_deps(data):
    object_ids = []
    network_id = data.get('network_id')
    if network_id is not None:
        object_ids.append(network_id)

    gateway_id = data.get('gateway_id')
    if gateway_id is not None:
        object_ids.append(gateway_id)

    return object_ids


def _generate_sfc_port_pair_deps(data):
    object_ids = []
    ingress_port = data.get('ingress')
    if ingress_port is not None:
        object_ids.append(ingress_port)

    egress_port = data.get('egress')
    if egress_port is not None:
        object_ids.append(egress_port)

    return object_ids


def _generate_sfc_port_pair_group_deps(data):
    return data['port_pairs']


def _generate_sfc_port_chain_deps(data):
    object_ids = data['port_pair_groups'][:]
    flow_classifiers = data['flow_classifiers'][:]
    object_ids.extend(flow_classifiers)

    return object_ids


def _generate_bgpvpn_deps(data):
    object_ids = []

    network_ids = data.get('networks')
    if network_ids is not None:
        object_ids.extend(network_ids)

    router_ids = data.get('routers')
    if router_ids is not None:
        object_ids.extend(router_ids)

    return object_ids


_CREATE_OR_UPDATE_DEP_GENERATOR = {
    odl_const.ODL_NETWORK: _generate_network_deps,
    odl_const.ODL_SUBNET: _generate_subnet_deps,
    odl_const.ODL_PORT: _generate_port_deps,
    # TODO(yamahata): dependency between SG and PORT
    odl_const.ODL_SG_RULE: _generate_sg_rule_deps,
    odl_const.ODL_ROUTER: _generate_router_deps,
    odl_const.ODL_FLOATINGIP: _generate_floatingip_deps,
    odl_const.ODL_TRUNK: _generate_trunk_deps,
    odl_const.ODL_L2GATEWAY_CONNECTION: _generate_l2gateway_connection_deps,
    odl_const.ODL_SFC_PORT_PAIR: _generate_sfc_port_pair_deps,
    odl_const.ODL_SFC_PORT_PAIR_GROUP: _generate_sfc_port_pair_group_deps,
    odl_const.ODL_SFC_PORT_CHAIN: _generate_sfc_port_chain_deps,
    odl_const.ODL_BGPVPN: _generate_bgpvpn_deps,
}


_DELETE_DEPENDENCIES = {
    odl_const.ODL_NETWORK: (odl_const.ODL_SUBNET, odl_const.ODL_PORT,
                            odl_const.ODL_ROUTER,
                            odl_const.ODL_L2GATEWAY_CONNECTION,
                            odl_const.ODL_BGPVPN),
    odl_const.ODL_SUBNET: (odl_const.ODL_PORT,),
    odl_const.ODL_ROUTER: (odl_const.ODL_PORT, odl_const.ODL_FLOATINGIP,
                           odl_const.ODL_BGPVPN),
    odl_const.ODL_PORT: (odl_const.ODL_TRUNK,),
    # TODO(yamahata): dependency between SG and PORT
    odl_const.ODL_SG: (odl_const.ODL_SG_RULE,),
    odl_const.ODL_L2GATEWAY: (odl_const.ODL_L2GATEWAY_CONNECTION,),
    odl_const.ODL_SFC_FLOW_CLASSIFIER: (odl_const.ODL_SFC_PORT_CHAIN,),
    odl_const.ODL_SFC_PORT_PAIR: (odl_const.ODL_SFC_PORT_PAIR_GROUP,),
    odl_const.ODL_SFC_PORT_PAIR_GROUP: (odl_const.ODL_SFC_PORT_CHAIN,),
    odl_const.ODL_QOS_POLICY: (odl_const.ODL_PORT, odl_const.ODL_NETWORK),
}


def calculate(context, operation, object_type, object_uuid, data):
    """Calculate resource deps in journaled operations.

    As a rule of thumb validation takes into consideration only operations in
    pending or processing state, other states are irrelevant.
    :param context: enginefacade context
    :param row: entry in journal entry to be validated
    """
    deps = []
    if operation == odl_const.ODL_DELETE:
        return _get_delete_dependencies(context, object_type, object_uuid)
    elif operation == odl_const.ODL_UPDATE:
        deps.extend(
            db.get_pending_or_processing_ops(
                context, object_uuid,
                operation=(odl_const.ODL_CREATE, odl_const.ODL_UPDATE)))
    elif operation != odl_const.ODL_CREATE:
        raise ValueError(_("unsupported operation {}").format(operation))

    # Validate deps if there are any to validate.
    dep_generator = _CREATE_OR_UPDATE_DEP_GENERATOR.get(object_type)
    if dep_generator is not None:
        object_ids = dep_generator(data)
        if object_ids is not None:
            deps.extend(_get_older_operations(context, object_ids))

    return deps



# Copyright (c) 2017 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import threading

from neutron_lib.callbacks import resources
from neutron_lib import constants as n_const
from neutron_lib import context
from neutron_lib.plugins import directory
from neutron_lib import worker
from oslo_log import log

from neutron.db import provisioning_blocks

from networking_odl.common import client as odl_client
from networking_odl.common import odl_features
from networking_odl.common import utils
from networking_odl.common import websocket_client as odl_ws_client

LOG = log.getLogger(__name__)


class OdlPortStatusUpdate(worker.BaseWorker):
    """Class to register and handle port status update"""
    PORT_PATH = "restconf/operational/neutron:neutron/ports/port"

    def __init__(self):
        super(OdlPortStatusUpdate, self).__init__()
        self.odl_websocket_client = None

    def start(self):
        super(OdlPortStatusUpdate, self).start()
        LOG.debug('OdlPortStatusUpdate worker running')
        if odl_features.has(odl_features.OPERATIONAL_PORT_STATUS):
            self.run_websocket()

    def stop(self):
        if self.odl_websocket_client:
            self.odl_websocket_client.set_exit_flag()

    def wait(self):
        """Wait for service to complete."""

    @staticmethod
    def reset():
        pass

    def run_websocket(self):
        # OpenDaylight path to recieve websocket notifications on
        neutron_ports_path = "/neutron:neutron/neutron:ports"

        self.path_uri = utils.get_odl_url()

        self.odl_websocket_client = (
            odl_ws_client.OpenDaylightWebsocketClient.odl_create_websocket(
                self.path_uri, neutron_ports_path,
                odl_ws_client.ODL_OPERATIONAL_DATASTORE,
                odl_ws_client.ODL_NOTIFICATION_SCOPE_SUBTREE,
                self._process_websocket_recv,
                self._process_websocket_reconnect,
                True
            ))

    def _process_websocket_recv(self, payload, reconnect):
        # Callback for websocket notification
        LOG.debug("Websocket notification for port status update")
        for event in odl_ws_client.EventDataParser.get_item(payload):
            operation, path, data = event.get_fields()
            if ((operation in [event.OPERATION_UPDATE,
                 event.OPERATION_CREATE])):
                port_id = event.extract_field(path, "neutron:uuid")
                port_id = str(port_id).strip("'")
                status_field = data.get('status')
                if status_field is not None:
                    status = status_field.get('content')
                    LOG.debug("Update port for port id %s %s", port_id, status)
                    # for now we only support transition from DOWN->ACTIVE
                    # https://bugs.launchpad.net/networking-odl/+bug/1686023
                    if status == n_const.PORT_STATUS_ACTIVE:
                        provisioning_blocks.provisioning_complete(
                            context.get_admin_context(),
                            port_id, resources.PORT,
                            provisioning_blocks.L2_AGENT_ENTITY)
            if operation == event.OPERATION_DELETE:
                LOG.debug("PortStatus: Ignoring delete operation")

    def _process_websocket_reconnect(self, status):
        if status == odl_ws_client.ODL_WEBSOCKET_CONNECTED:
            # Get port data using restconf
            LOG.debug("Websocket notification on reconnection")
            reconn_thread = threading.Thread(
                name='websocket', target=self._pull_missed_statuses)
            reconn_thread.start()

    def _pull_missed_statuses(self):
        LOG.debug("starting to pull pending statuses...")
        plugin = directory.get_plugin()
        filter = {"status": [n_const.PORT_STATUS_DOWN],
                  "vif_type": ["unbound"]}
        ports = plugin.get_ports(context.get_admin_context(), filter)

        if not ports:
            LOG.debug("no down ports found, done")
            return

        port_fetch_url = utils.get_odl_url(self.PORT_PATH)
        client = odl_client.OpenDaylightRestClient.create_client(
            url=port_fetch_url)

        for port in ports:
            port_id = port["id"]
            response = client.get(port_id)
            if response.status_code != 200:
                LOG.warning("Non-200 response code %s", str(response))
                continue
            odl_status = response.json()['port'][0]['status']
            if odl_status == n_const.PORT_STATUS_ACTIVE:
                # for now we only support transition from DOWN->ACTIVE
                # See https://bugs.launchpad.net/networking-odl/+bug/1686023
                provisioning_blocks.provisioning_complete(
                    context.get_admin_context(),
                    port_id, resources.PORT,
                    provisioning_blocks.L2_AGENT_ENTITY)
        LOG.debug("done pulling pending statuses")



#  Copyright (c) 2016 OpenStack Foundation
#  All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.
#

from neutron.db import extraroute_db
from neutron.db import l3_agentschedulers_db
from neutron.db import l3_dvr_db
from neutron.db import l3_gwmode_db
from neutron_lib import constants as q_const
from neutron_lib.db import api as db_api
from neutron_lib.plugins import constants as plugin_constants
from oslo_log import log as logging


from networking_odl.common import config  # noqa
from networking_odl.common import constants as odl_const
from networking_odl.journal import full_sync
from networking_odl.journal import journal

LOG = logging.getLogger(__name__)

L3_RESOURCES = {
    odl_const.ODL_ROUTER: odl_const.ODL_ROUTERS,
    odl_const.ODL_FLOATINGIP: odl_const.ODL_FLOATINGIPS
}


@db_api.retry_if_session_inactive()
@db_api.CONTEXT_WRITER.savepoint
def _record_in_journal(context, object_type, operation, object_id, data):
    journal.record(context, object_type, object_id, operation, data)


class OpenDaylightL3RouterPlugin(
        extraroute_db.ExtraRoute_db_mixin,
        l3_dvr_db.L3_NAT_with_dvr_db_mixin,
        l3_gwmode_db.L3_NAT_db_mixin,
        l3_agentschedulers_db.L3AgentSchedulerDbMixin):

    """Implementation of the OpenDaylight L3 Router Service Plugin.

    This class implements a L3 service plugin that provides
    router and floatingip resources and manages associated
    request/response.
    """
    supported_extension_aliases = ["dvr", "router", "ext-gw-mode",
                                   "extraroute"]

    def __init__(self):
        super(OpenDaylightL3RouterPlugin, self).__init__()
        self.journal = journal.OpenDaylightJournalThread()
        full_sync.register(plugin_constants.L3, L3_RESOURCES)

    def get_plugin_type(self):
        return plugin_constants.L3

    def get_plugin_description(self):
        """Returns string description of the plugin."""
        return ("L3 Router Service Plugin for basic L3 forwarding "
                "using OpenDaylight.")

    @journal.call_thread_on_end
    def create_router(self, context, router):
        router_dict = super(
            OpenDaylightL3RouterPlugin, self).create_router(context, router)
        _record_in_journal(
            context, odl_const.ODL_ROUTER, odl_const.ODL_CREATE,
            router_dict['id'], router_dict)
        return router_dict

    @journal.call_thread_on_end
    def update_router(self, context, id, router):
        router_dict = super(
            OpenDaylightL3RouterPlugin, self).update_router(
                context, id, router)
        _record_in_journal(
            context, odl_const.ODL_ROUTER, odl_const.ODL_UPDATE,
            id, router_dict)
        return router_dict

    @journal.call_thread_on_end
    def delete_router(self, context, id):
        router_dict = self.get_router(context, id)
        dependency_list = [router_dict['gw_port_id']]
        super(OpenDaylightL3RouterPlugin, self).delete_router(context,
                                                              id)
        _record_in_journal(
            context, odl_const.ODL_ROUTER, odl_const.ODL_DELETE,
            id, dependency_list)

    @journal.call_thread_on_end
    def create_floatingip(self, context, floatingip,
                          initial_status=q_const.FLOATINGIP_STATUS_ACTIVE):
        fip = floatingip['floatingip']
        if fip.get('port_id') is None:
            initial_status = q_const.FLOATINGIP_STATUS_DOWN
        fip_dict = super(
            OpenDaylightL3RouterPlugin, self).create_floatingip(
                context, floatingip, initial_status)
        _record_in_journal(
            context, odl_const.ODL_FLOATINGIP, odl_const.ODL_CREATE,
            fip_dict['id'], fip_dict)
        return fip_dict

    @journal.call_thread_on_end
    def update_floatingip(self, context, id, floatingip):
        fip_dict = super(
            OpenDaylightL3RouterPlugin, self).update_floatingip(
                context, id, floatingip)

        # Update status based on association
        if fip_dict.get('port_id') is None:
            fip_dict['status'] = q_const.FLOATINGIP_STATUS_DOWN
        else:
            fip_dict['status'] = q_const.FLOATINGIP_STATUS_ACTIVE
        self.update_floatingip_status(context, id,
                                      fip_dict['status'])

        _record_in_journal(
            context, odl_const.ODL_FLOATINGIP, odl_const.ODL_UPDATE,
            id, fip_dict)
        return fip_dict

    @journal.call_thread_on_end
    def delete_floatingip(self, context, id):
        floatingip_dict = self.get_floatingip(context, id)
        dependency_list = [floatingip_dict['router_id'],
                           floatingip_dict['floating_network_id']]
        super(OpenDaylightL3RouterPlugin, self).delete_floatingip(
            context, id)
        _record_in_journal(
            context, odl_const.ODL_FLOATINGIP, odl_const.ODL_DELETE,
            id, dependency_list)

    def disassociate_floatingips(self, context, port_id, do_notify=True):
        fip_dicts = self.get_floatingips(context,
                                         filters={'port_id': [port_id]})
        router_ids = super(
            OpenDaylightL3RouterPlugin, self).disassociate_floatingips(
                context, port_id, do_notify)
        for fip_dict in fip_dicts:
            fip_dict = self.get_floatingip(context, fip_dict['id'])
            fip_dict['status'] = q_const.FLOATINGIP_STATUS_DOWN
            self.update_floatingip_status(context, fip_dict['id'],
                                          fip_dict['status'])
            _record_in_journal(
                context, odl_const.ODL_FLOATINGIP, odl_const.ODL_UPDATE,
                fip_dict['id'], fip_dict)
        return router_ids

    @journal.call_thread_on_end
    def add_router_interface(self, context, router_id, interface_info):
        new_router = super(
            OpenDaylightL3RouterPlugin, self).add_router_interface(
                context, router_id, interface_info)
        return new_router

    @journal.call_thread_on_end
    def remove_router_interface(self, context, router_id, interface_info):
        new_router = super(
            OpenDaylightL3RouterPlugin, self).remove_router_interface(
                context, router_id, interface_info)
        return new_router

    dvr_deletens_if_no_port_warned = False

    def dvr_deletens_if_no_port(self, context, port_id):
        # TODO(yamahata): implement this method or delete this logging
        # For now, this is defined to avoid attribute exception
        # Since ODL L3 does not create namespaces, this is always going to
        # be a noop. When it is confirmed, delete this comment and logging
        if not self.dvr_deletens_if_no_port_warned:
            LOG.debug('dvr is not suported yet. '
                      'this method needs to be implemented')
            self.dvr_deletens_if_no_port_warned = True
        return []



# Copyright (c) 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

ODL_NETWORK = 'network'
ODL_NETWORKS = 'networks'
ODL_SUBNET = 'subnet'
ODL_SUBNETS = 'subnets'
ODL_PORT = 'port'
ODL_PORTS = 'ports'
ODL_SG = 'security_group'
ODL_SGS = 'security_groups'
ODL_SG_RULE = 'security_group_rule'
ODL_SG_RULES = 'security_group_rules'
ODL_ROUTER = 'router'
ODL_ROUTERS = 'routers'
ODL_FLOATINGIP = 'floatingip'
ODL_FLOATINGIPS = 'floatingips'

ODL_QOS = 'qos'
ODL_QOS_POLICY = 'policy'
ODL_QOS_POLICIES = 'policies'

ODL_SFC = 'sfc'
ODL_SFC_FLOW_CLASSIFIER = 'flowclassifier'
ODL_SFC_FLOW_CLASSIFIERS = 'flowclassifiers'
ODL_SFC_PORT_PAIR = 'portpair'
ODL_SFC_PORT_PAIRS = 'portpairs'
ODL_SFC_PORT_PAIR_GROUP = 'portpairgroup'
ODL_SFC_PORT_PAIR_GROUPS = 'portpairgroups'
ODL_SFC_PORT_CHAIN = 'portchain'
ODL_SFC_PORT_CHAINS = 'portchains'

NETWORKING_SFC_FLOW_CLASSIFIER = 'flow_classifier'
NETWORKING_SFC_FLOW_CLASSIFIERS = 'flow_classifiers'
NETWORKING_SFC_PORT_PAIR = 'port_pair'
NETWORKING_SFC_PORT_PAIRS = 'port_pairs'
NETWORKING_SFC_PORT_PAIR_GROUP = 'port_pair_group'
NETWORKING_SFC_PORT_PAIR_GROUPS = 'port_pair_groups'
NETWORKING_SFC_PORT_CHAIN = 'port_chain'
NETWORKING_SFC_PORT_CHAINS = 'port_chains'

ODL_TRUNK = 'trunk'
ODL_TRUNKS = 'trunks'

ODL_L2GATEWAY = 'l2_gateway'
ODL_L2GATEWAYS = 'l2_gateways'
ODL_L2GATEWAY_CONNECTION = 'l2gateway_connection'
ODL_L2GATEWAY_CONNECTIONS = 'l2_gateway_connections'

ODL_BGPVPN = 'bgpvpn'
ODL_BGPVPNS = 'bgpvpns'
ODL_BGPVPN_NETWORK_ASSOCIATION = 'bgpvpn_network_association'
ODL_BGPVPN_NETWORK_ASSOCIATIONS = 'bgpvpn_network_associations'
ODL_BGPVPN_ROUTER_ASSOCIATION = 'bgpvpn_router_association'
ODL_BGPVPN_ROUTER_ASSOCIATIONS = 'bgpvpn_router_associations'

ODL_ML2_MECH_DRIVER_V2 = "opendaylight_v2"

ODL_DEVICE_ID_START = 'OpenDaylight'

ODL_CREATE = 'create'
ODL_UPDATE = 'update'
ODL_DELETE = 'delete'

# Constants for journal operation states
PENDING = 'pending'
PROCESSING = 'processing'
FAILED = 'failed'
COMPLETED = 'completed'

# Journal Callback events
BEFORE_COMPLETE = 'before_complete'
AFTER_DELETE = 'after_delete'

# dict to store url mappings
RESOURCE_URL_MAPPINGS = {
    ODL_QOS_POLICY: "%s/%s" % (ODL_QOS, ODL_QOS_POLICIES),
    ODL_SFC_FLOW_CLASSIFIER: "%s/%s" % (ODL_SFC, ODL_SFC_FLOW_CLASSIFIERS),
    ODL_SFC_PORT_CHAIN: "%s/%s" % (ODL_SFC, ODL_SFC_PORT_CHAINS),
    ODL_SFC_PORT_PAIR: "%s/%s" % (ODL_SFC, ODL_SFC_PORT_PAIRS),
    ODL_SFC_PORT_PAIR_GROUP: "%s/%s" % (ODL_SFC, ODL_SFC_PORT_PAIR_GROUPS)
}



# Copyright (c) 2016 Brocade Communication Systems
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

CLASSIFIERS_BASE_URI = 'sfc/flowclassifiers'
FAKE_FLOW_CLASSIFIER_ID = "4a334cd4-fe9c-4fae-af4b-321c5e2eb051"
FAKE_FLOW_CLASSIFIER = {
    "id": "4a334cd4-fe9c-4fae-af4b-321c5e2eb051",
    "name": "FC1",
    "tenant_id": "1814726e2d22407b8ca76db5e567dcf1",
    "description": "Flow rule for classifying TCP traffic",
    "protocol": "TCP",
    "source_port_range_min": 22,
    "source_port_range_max": 4000,
    "destination_port_range_min": 80,
    "destination_port_range_max": 80,
    "source_ip_prefix": "22.12.34.44",
    "destination_ip_prefix": "22.12.34.45"
}
PORT_PAIRS_BASE_URI = 'sfc/portpairs'
FAKE_PORT_PAIR_ID = "78dcd363-fc23-aeb6-f44b-56dc5e2fb3ae"
FAKE_PORT_PAIR = {
    "name": "SF1",
    "id": "78dcd363-fc23-aeb6-f44b-56dc5e2fb3ae",
    "tenant_id": "d382007aa9904763a801f68ecf065cf5",
    "description": "Firewall SF instance",
    "ingress": "dace4513-24fc-4fae-af4b-321c5e2eb3d1",
    "egress": "aef3478a-4a56-2a6e-cd3a-9dee4e2ec345"
}
PORT_PAIR_GROUPS_BASE_URI = 'sfc/portpairgroups'
FAKE_PORT_PAIR_GROUP_ID = "4512d643-24fc-4fae-af4b-321c5e2eb3d1"
FAKE_PORT_PAIR_GROUP = {
    "name": "Firewall_PortPairGroup",
    "id": "4512d643-24fc-4fae-af4b-321c5e2eb3d1",
    "tenant_id": "d382007aa9904763a801f68ecf065cf5",
    "description": "Grouping Firewall SF instances",
    "port_pairs": ["78dcd363-fc23-aeb6-f44b-56dc5e2fb3ae"]
}
PORT_CHAINS_BASE_URI = 'sfc/portchains'
FAKE_PORT_CHAIN_ID = "1278dcd4-459f-62ed-754b-87fc5e4a6751"
FAKE_PORT_CHAIN = {
    "name": "PC2",
    "id": "1278dcd4-459f-62ed-754b-87fc5e4a6751",
    "tenant_id": "d382007aa9904763a801f68ecf065cf5",
    "description": "Steering TCP and UDP traffic first to Firewall "
                   "and then to Loadbalancer",
    "flow_classifiers": ["4a334cd4-fe9c-4fae-af4b-321c5e2eb051",
                         "105a4b0a-73d6-11e5-b392-2c27d72acb4c"],
    "port_pair_groups": ["4512d643-24fc-4fae-af4b-321c5e2eb3d1",
                         "4a634d49-76dc-4fae-af4b-321c5e23d651"]
}



#
# Copyright (C) 2017 Ericsson India Global Services Pvt Ltd.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.
#

import contextlib
import copy
from unittest import mock

import webob.exc

from neutron.api import extensions as api_extensions
from neutron.db import servicetype_db as sdb
from neutron.tests import base as test_base
from neutron.tests.unit.plugins.ml2 import test_plugin
from oslo_utils import uuidutils

from networking_l2gw import extensions as l2gw_extensions
from networking_l2gw.services.l2gateway.common import constants as l2gw_const
from networking_l2gw.services.l2gateway.plugin import L2GatewayPlugin
from networking_odl.common import constants as odl_const
from networking_odl.tests.functional import base

_uuid = uuidutils.generate_uuid


class L2GatewayTestCaseMixin(object):

    devices = [{'device_name': 's1',
                'interfaces': [{'name': 'int1'}]
                },
               {'device_name': 's2',
                'interfaces': [{'name': 'int2', 'segmentation_id': [10, 20]}]
                }]
    l2gw_data = {l2gw_const.GATEWAY_RESOURCE_NAME: {'tenant_id': _uuid(),
                                                    'name': 'l2gw',
                                                    'devices': devices}}

    def setUp(self):
        """Perform parent setup with the common plugin configuration class."""
        # Ensure that the parent setup can be called without arguments
        # by the common configuration setUp.
        bits = self.service_provider.split(':')
        provider = {
            'service_type': bits[0],
            'name': bits[1],
            'driver': bits[2],
            'default': True
        }

        # override the default service provider
        self.service_providers = (
            mock.patch.object(sdb.ServiceTypeManager,
                              'get_service_providers').start())
        self.service_providers.return_value = [provider]
        super(L2GatewayTestCaseMixin, self).setUp()

    @contextlib.contextmanager
    def l2gateway(self, do_delete=True, **kwargs):
        req_data = copy.deepcopy(self.l2gw_data)

        fmt = 'json'
        if kwargs.get('data'):
            req_data = kwargs.get('data')
        else:
            req_data[l2gw_const.GATEWAY_RESOURCE_NAME].update(kwargs)
        l2gw_req = self.new_create_request(l2gw_const.L2_GATEWAYS,
                                           req_data, fmt=fmt)
        res = l2gw_req.get_response(self.ext_api)
        if res.status_int >= 400:
            raise webob.exc.HTTPClientError(code=res.status_int)
        l2gw = self.deserialize('json', res)
        yield l2gw
        if do_delete:
            self._delete(l2gw_const.L2_GATEWAYS,
                         l2gw[l2gw_const.GATEWAY_RESOURCE_NAME]['id'])

    @contextlib.contextmanager
    def l2gateway_connection(self, nw_id, l2gw_id,
                             do_delete=True, **kwargs):
        req_data = {
            l2gw_const.CONNECTION_RESOURCE_NAME:
                {'tenant_id': _uuid(),
                 'network_id': nw_id,
                 'l2_gateway_id': l2gw_id}
        }

        fmt = 'json'
        if kwargs.get('data'):
            req_data = kwargs.get('data')
        else:
            req_data[l2gw_const.CONNECTION_RESOURCE_NAME].update(kwargs)
        l2gw_connection_req = self.new_create_request(
            l2gw_const.L2_GATEWAYS_CONNECTION, req_data, fmt=fmt)
        res = l2gw_connection_req.get_response(self.ext_api)
        if res.status_int >= 400:
            raise webob.exc.HTTPClientError(code=res.status_int)
        l2gw_connection = self.deserialize('json', res)
        yield l2gw_connection
        if do_delete:
            self._delete(l2gw_const.L2_GATEWAYS_CONNECTION,
                         l2gw_connection
                         [l2gw_const.CONNECTION_RESOURCE_NAME]['id'])

    @staticmethod
    def convert_to_odl_l2gw_connection(l2gw_connection_in):
        odl_l2_gw_conn_data = copy.deepcopy(
            l2gw_connection_in[l2gw_const.CONNECTION_RESOURCE_NAME])
        odl_l2_gw_conn_data['gateway_id'] = (
            odl_l2_gw_conn_data['l2_gateway_id'])
        odl_l2_gw_conn_data.pop('l2_gateway_id')
        return {odl_const.ODL_L2GATEWAY_CONNECTION: odl_l2_gw_conn_data}


class _TestL2GatewayBase(base.OdlTestsBase, L2GatewayTestCaseMixin):

    def get_ext_managers(self):
        extensions_path = ':'.join(l2gw_extensions.__path__)
        return api_extensions.PluginAwareExtensionManager(
            extensions_path,
            {'l2gw_plugin': L2GatewayPlugin()})

    def get_additional_service_plugins(self):
        l2gw_plugin_str = ('networking_l2gw.services.l2gateway.plugin.'
                           'L2GatewayPlugin')
        service_plugin = {'l2gw_plugin': l2gw_plugin_str}
        return service_plugin

    @test_base.unstable_test('bug/1989718')
    def test_l2gateway_create(self):
        with self.l2gateway(name='mygateway') as l2gateway:
            self.assert_resource_created(odl_const.ODL_L2GATEWAY, l2gateway)

    @test_base.unstable_test('bug/1989718')
    def test_l2gateway_update(self):
        with self.l2gateway(name='gateway1') as l2gateway:
            self.resource_update_test(odl_const.ODL_L2GATEWAY, l2gateway)

    @test_base.unstable_test('bug/1989718')
    def test_l2gateway_delete(self):
        with self.l2gateway(do_delete=False) as l2gateway:
            self.resource_delete_test(odl_const.ODL_L2GATEWAY, l2gateway)

    @test_base.unstable_test('bug/1989718')
    def test_l2gateway_connection_create_delete(self):
        odl_l2gw_connection = {}
        with self.network() as network:
            with self.l2gateway() as l2gateway:
                net_id = network['network']['id']
                l2gw_id = l2gateway[odl_const.ODL_L2GATEWAY]['id']
                with (self.l2gateway_connection(net_id, l2gw_id)
                      ) as l2gw_connection:
                    odl_l2gw_connection = (
                        self.convert_to_odl_l2gw_connection(l2gw_connection))
                    self.assert_resource_created(
                        odl_const.ODL_L2GATEWAY_CONNECTION,
                        odl_l2gw_connection)
                self.assertIsNone(self.get_odl_resource(
                    odl_const.ODL_L2GATEWAY_CONNECTION, odl_l2gw_connection))


class TestL2gatewayV2Driver(base.V2DriverAdjustment, _TestL2GatewayBase,
                            test_plugin.Ml2PluginV2TestCase):
    _mechanism_drivers = ['opendaylight_v2']
    service_provider = ('L2GW:OpenDaylight:networking_odl.l2gateway.driver_v2.'
                        'OpenDaylightL2gwDriver:default')



# Copyright 2016 Isaku Yamahata <isaku.yamahata at gmail.com>
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""update opendayligut journal

Revision ID: fa0c536252a5
Revises: 383acb0d38a0
Create Date: 2016-08-05 23:03:46.470595

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = 'fa0c536252a5'
down_revision = '383acb0d38a0'
depends_on = ('3d560427d776', )


def upgrade():
    # Since a new primary key is introduced and alembic doesn't allow to
    # add new primary key, create a new table with new primary key and
    # rename it.
    op.execute("INSERT INTO opendaylightjournal_new "
               "(object_type, object_uuid, operation, data, "
               "state, retry_count, created_at, last_retried) "
               "SELECT object_type, object_uuid, operation, data, "
               "state, retry_count, created_at, last_retried "
               "FROM opendaylightjournal "
               "WHERE state != 'completed' "
               "ORDER BY created_at ASC")
    op.drop_table('opendaylightjournal')
    op.rename_table('opendaylightjournal_new', 'opendaylightjournal')



#
# Copyright (C) 2016 Red Hat, Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.
#

from neutron_lib import context as neutron_context
from neutron_lib.db import api as db_api
from oslo_log import log as logging
from oslo_service import loopingcall


from networking_odl.db import db


LOG = logging.getLogger(__name__)


class PeriodicTask(object):
    def __init__(self, task, interval):
        self.task = task
        self.phases = []
        self.timer = loopingcall.FixedIntervalLoopingCall(self.execute_ops)
        self.interval = interval

    def start(self):
        self.timer.start(self.interval, stop_on_exception=False)

    def cleanup(self):
        # this method is used for unit test to tear down
        self.timer.stop()
        try:
            self.timer.wait()
        except AttributeError:
            # NOTE(yamahata): workaround
            # some tests call this cleanup without calling start
            pass

    @db_api.retry_if_session_inactive()
    @db_api.CONTEXT_WRITER.savepoint
    def _set_operation(self, context, operation):
        db.update_periodic_task(context, task=self.task,
                                operation=operation)

    def _execute_op(self, operation, context):
        op_details = operation.__name__
        if operation.__doc__:
            op_details += " (%s)" % operation.__doc__

        try:
            LOG.info("Starting %s phase of periodic task %s.",
                     op_details, self.task)
            self._set_operation(context, operation)
            operation(context)
            LOG.info("Finished %s phase of %s task.", op_details, self.task)
        except Exception:
            LOG.exception("Failed during periodic task operation %s.",
                          op_details)

    def task_already_executed_recently(self, context):
        return db.was_periodic_task_executed_recently(
            context, self.task, self.interval)

    @db_api.retry_if_session_inactive()
    @db_api.CONTEXT_WRITER.savepoint
    def _clear_and_unlock_task(self, context):
        db.update_periodic_task(context, task=self.task,
                                operation=None)
        db.unlock_periodic_task(context, self.task)

    @db_api.retry_if_session_inactive()
    @db_api.CONTEXT_WRITER.savepoint
    def _lock_task(self, context):
        return db.lock_periodic_task(context, self.task)

    def execute_ops(self, forced=False):
        LOG.info("Starting %s periodic task.", self.task)
        context = neutron_context.get_admin_context()

        # Lock make sure that periodic task is executed only after
        # specified interval. It makes sure that maintenance tasks
        # are not executed back to back.
        if not forced and self.task_already_executed_recently(context):
            LOG.info("Periodic %s task executed after periodic interval "
                     "Skipping execution.", self.task)
            return

        if not self._lock_task(context):
            LOG.info("Periodic %s task already running task", self.task)
            return

        try:
            for phase in self.phases:
                self._execute_op(phase, context)
        finally:
            self._clear_and_unlock_task(context)

        LOG.info("%s task has been finished", self.task)

    def register_operation(self, phase):
        """Register a function to be run by the periodic task.

        :param phase: Function to call when the thread runs. The function will
        receive a DB session to use for DB operations.
        """
        self.phases.append(phase)
        LOG.info("%s phase has been registered in %s task", phase, self.task)



# Copyright 2017 NEC Corp
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""create opendaylight_preiodic_task table

Revision ID: 6f7dfb241354
Revises: 0472f56ff2fb
Create Date: 2017-05-24 03:01:00.755796

"""

from alembic import op
import sqlalchemy as sa

from networking_odl.common import constants as odl_const

# revision identifiers, used by Alembic.
revision = '6f7dfb241354'
down_revision = '0472f56ff2fb'


def upgrade():
    periodic_table = op.create_table(
        'opendaylight_periodic_task',
        sa.Column('state', sa.Enum(odl_const.PENDING, odl_const.PROCESSING,
                                   name='state'),
                  nullable=False),
        sa.Column('processing_operation', sa.String(70)),
        sa.Column('task', sa.String(70), primary_key=True),
        sa.Column('lock_updated', sa.TIMESTAMP, nullable=False,
                  server_default=sa.func.now(),
                  onupdate=sa.func.now())
    )
    op.bulk_insert(periodic_table,
                   [{'task': 'maintenance',
                     'state': odl_const.PENDING},
                    {'task': 'hostconfig',
                     'state': odl_const.PENDING}])



# Copyright (c) 2017 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re
import threading
from unittest import mock


from networking_odl.common.client import OpenDaylightRestClient
from networking_odl.common import websocket_client as odl_ws_client
from networking_odl.common.websocket_client import OpenDaylightWebsocketClient
from networking_odl.ml2.port_status_update import OdlPortStatusUpdate
from networking_odl.tests import base
from neutron.db import provisioning_blocks
import neutron_lib.context
import neutron_lib.plugins.directory


class TestOdlPortStatusUpdate(base.DietTestCase):

    WEBSOCK_NOTIFICATION = re.sub(r'\s*', '', """
        {
            "notification": {
                "data-changed-notification": {
                    "data-change-event": {
                        "data": {
                            "status": {
                                "content": "ACTIVE",
                                "xmlns": "urn:opendaylight:neutron"
                            }
                        },
                        "operation": "updated",
                        "path":
                        "/neutron:neutron/neutron:ports/neutron:port[
                        neutron:uuid='d6e6335d-9568-4949-aef1-4107e34c5f28']
                        /neutron:status"
                    },
                    "xmlns":
                "urn:opendaylight:params:xml:ns:yang:controller:md:sal:remote"
                },
                "eventTime": "2017-02-22T02:27:32+02:00",
                "xmlns": "urn:ietf:params:xml:ns:netconf:notification:1.0"
            }
        }""")

    def setUp(self):
        self.useFixture(base.OpenDaylightFeaturesFixture())
        self.mock_ws_client = mock.patch.object(
            OpenDaylightWebsocketClient, 'odl_create_websocket')
        super(TestOdlPortStatusUpdate, self).setUp()

    def test_object_create(self):
        OdlPortStatusUpdate()

    @mock.patch.object(provisioning_blocks, 'provisioning_complete')
    def test_websock_recv(self, mocked_provisioning_complete):
        updater = OdlPortStatusUpdate()
        updater._process_websocket_recv(self.WEBSOCK_NOTIFICATION, False)
        mocked_provisioning_complete.assert_called_once()
        self.assertEqual(mocked_provisioning_complete.call_args[0][1],
                         'd6e6335d-9568-4949-aef1-4107e34c5f28')

    @mock.patch.object(provisioning_blocks, 'provisioning_complete')
    @mock.patch.object(neutron_lib.context, 'get_admin_context')
    @mock.patch.object(OpenDaylightRestClient, 'get')
    @mock.patch.object(neutron_lib.plugins.directory, 'get_plugin')
    def test_pull_missed_statuses(self, mocked_get_plugin, mocked_get, ac, pc):
        uuid = 'd6e6335d-9568-4949-aef1-4107e34c5f28'
        plugin = mock.MagicMock()
        plugin.get_ports = mock.MagicMock(return_value=[{'id': uuid}])
        mocked_get_plugin.return_value = plugin

        updater = OdlPortStatusUpdate()
        updater._pull_missed_statuses()

        mocked_get.assert_called_with(uuid)

    @mock.patch.object(threading, 'Thread')
    def test_process_websocket_reconnect(self, mocked_thread):
        updater = OdlPortStatusUpdate()
        updater._process_websocket_reconnect(
            odl_ws_client.ODL_WEBSOCKET_CONNECTED)
        mocked_thread.assert_called()
        mocked_thread.return_value.start.assert_called()



# Copyright 2017 NEC Corp
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""drop opendaylight_maintenance table

Revision ID: eccd865b7d3a
Revises: fa0c536252a5
Create Date: 2017-05-24 03:00:40.194278

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = 'eccd865b7d3a'
down_revision = 'fa0c536252a5'


def upgrade():
    op.drop_table('opendaylight_maintenance')
